from functools import partial
from typing import Optional, Iterable
from uuid import UUID

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._rest.pageable import Pageable
from citrine._rest.paginator import Paginator
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.informatics.generative_design import GenerativeDesignResult


class GenerativeDesignExecution(
    Resource["GenerativeDesignExecution"], Pageable, AsynchronousObject
):
    """The execution of a Generative Design task.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Design executions also have a ``status_description`` field with more information.

    """

    _paginator: Paginator = Paginator()
    _collection_key = 'response'
    _in_progress_statuses = ["INPROGRESS"]
    _succeeded_statuses = ["SUCCEEDED"]
    _failed_statuses = ["FAILED"]
    _session: Optional[Session] = None
    project_id: Optional[UUID] = None

    uid: UUID = properties.UUID('id', serializable=False)
    """:UUID: Unique identifier of the execution"""
    status = properties.Optional(properties.String(), 'status', serializable=False)
    """:Optional[str]: short description of the execution's status"""
    status_description = properties.Optional(
        properties.String(), 'status_description', serializable=False)
    """:Optional[str]: more detailed description of the execution's status"""
    status_info = properties.Optional(
        properties.List(properties.String()),
        'status_info',
        serializable=False
    )

    """:Optional[List[str]]: human-readable explanations of the status"""
    created_by = properties.Optional(properties.UUID, 'created_by', serializable=False)
    """:Optional[UUID]: id of the user who created the resource"""
    updated_by = properties.Optional(properties.UUID, 'updated_by', serializable=False)
    """:Optional[UUID]: id of the user who most recently updated the resource,
    if it has been updated"""
    create_time = properties.Optional(properties.Datetime, 'create_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was created"""
    update_time = properties.Optional(properties.Datetime, 'update_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was most recently updated,
    if it has been updated"""

    def __init__(self):
        """The GenerativeDesign executions are not directly instantiated by the user."""
        pass  # pragma: no cover

    def __str__(self):
        return '<GenerativeDesignExecution {!r}>'.format(str(self.uid))

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/generative-design/executions/',
            project_id=self.project_id,
        )

    @classmethod
    def _build_results(
        cls, subset_collection: Iterable[dict]
    ) -> Iterable[GenerativeDesignResult]:
        for generation_result in subset_collection:
            yield GenerativeDesignResult.build(generation_result)

    def results(
        self,
        *,
        page: Optional[int] = None,
        per_page: int = 100,
    ) -> Iterable[GenerativeDesignResult]:
        """Fetch the Generative Design Results for the particular execution, paginated."""
        path = self._path() + f'{self.uid}/results'
        fetcher = partial(self._fetch_page, path=path, fetch_func=self._session.get_resource)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_results,
                                        page=page,
                                        per_page=per_page)

    def result(
        self,
        *,
        result_id: UUID,
    ) -> GenerativeDesignResult:
        """Fetch a Generative Design Result for the particular UID."""
        path = self._path() + f'{self.uid}/results/{result_id}'
        data = self._session.get_resource(path, version=self._api_version)
        result = GenerativeDesignResult.build(data)
        return result



from typing import List, Mapping, Any

from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.design_spaces.design_space import DesignSpace
from citrine._rest.ai_resource_metadata import AIResourceMetadata

__all__ = ['EnumeratedDesignSpace']


class EnumeratedDesignSpace(Resource['EnumeratedDesignSpace'], DesignSpace, AIResourceMetadata):
    """An explicit enumeration of candidate materials to score.

    Enumerated design spaces are intended to capture small spaces with fewer than
    1000 values.  For larger spaces, use the DataSourceDesignSpace.

    Parameters
    ----------
    name:str
        the name of the design space
    description:str
        the description of the design space
    descriptors: list[Descriptor]
        the list of descriptors included in the candidates of the design space
    data: list[dict]
        list of dicts of the shape `{<descriptor_key>: <descriptor_value>}`
        where each dict corresponds to a candidate in the design space

    """

    _resource_type = ResourceTypeEnum.MODULE

    descriptors = properties.List(properties.Object(Descriptor), 'config.descriptors')
    data = properties.List(properties.Mapping(properties.String, properties.Raw), 'config.data')

    typ = properties.String('config.type', default='EnumeratedDesignSpace', deserializable=False)
    module_type = properties.String('module_type', default='DESIGN_SPACE')

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 descriptors: List[Descriptor],
                 data: List[Mapping[str, Any]]):
        self.name: str = name
        self.description: str = description
        self.descriptors: List[Descriptor] = descriptors
        self.data: List[Mapping[str, Any]] = data

    def _post_dump(self, data: dict) -> dict:
        data['display_name'] = data['config']['name']
        return data

    def __str__(self):
        return '<EnumeratedDesignSpace {!r}>'.format(self.name)



from deprecation import deprecated
from functools import lru_cache
from typing import List, Optional
from uuid import UUID

from citrine._rest.resource import Resource
from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.informatics.predictor_evaluation_result import PredictorEvaluationResult


class PredictorEvaluationExecution(Resource['PredictorEvaluationExecution'], AsynchronousObject):
    """The execution of a PredictorEvaluationWorkflow.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Predictor evaluation executions also have a ``status_description`` field with more information.

    """

    _session: Optional[Session] = None
    _in_progress_statuses = ["INPROGRESS"]
    _succeeded_statuses = ["SUCCEEDED"]
    _failed_statuses = ["FAILED"]
    project_id: Optional[UUID] = None
    """:Optional[UUID]: Unique ID of the project that contains this execution."""

    uid: UUID = properties.UUID('id', serializable=False)
    """:UUID: Unique identifier of the workflow execution"""
    evaluator_names = properties.List(properties.String, "evaluator_names", serializable=False)
    """:List[str]: names of the predictor evaluators that were executed. These are used
    when calling the ``results()`` method."""
    workflow_id = properties.UUID('workflow_id', serializable=False)
    """:UUID: Unique identifier of the workflow that was executed"""
    predictor_id = properties.UUID('predictor_id', serializable=False)
    predictor_version = properties.Integer('predictor_version', serializable=False)
    status = properties.Optional(properties.String(), 'status', serializable=False)
    """:Optional[str]: short description of the execution's status"""
    status_info = properties.Optional(
        properties.List(properties.String()),
        'status_info',
        serializable=False
    )
    """:Optional[List[str]]: human-readable explanations of the status"""

    def __init__(self):
        """Predictor evaluation executions are not directly instantiated by the user."""
        pass  # pragma: no cover

    def __str__(self):
        return '<PredictorEvaluationExecution {!r}>'.format(str(self.uid))

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/predictor-evaluation-executions/{execution_id}',
            project_id=self.project_id,
            execution_id=self.uid
        )

    @lru_cache()
    def results(self, evaluator_name: str) -> PredictorEvaluationResult:
        """
        Get a specific evaluation result by the name of the evaluator that produced it.

        Parameters
        ----------
        evaluator_name: str
            Name of the evaluator for which to get the results

        Returns
        -------
        PredictorEvaluationResult
            The evaluation result from the evaluator with the given name

        """
        params = {"evaluator_name": evaluator_name}
        resource = self._session.get_resource(self._path() + "/results", params=params)
        return PredictorEvaluationResult.build(resource)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.results(item)
        else:
            raise TypeError("Results are accessed by string names")

    def __iter__(self):
        return iter(self.evaluator_names)

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental(self) -> bool:
        """[DEPRECATED] whether the execution is experimental (newer, less well-tested)."""  # noqa - insisting this docstring is a signature
        return False

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental_reasons(self) -> List[str]:
        """[DEPRECATED] human-readable reasons why the execution is experimental."""
        return []



"""Tools for working with Dimensions."""
from typing import Optional, Type, List

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.descriptors import Descriptor, RealDescriptor

__all__ = ['Dimension', 'ContinuousDimension', 'EnumeratedDimension']


class Dimension(PolymorphicSerializable['Dimension']):
    """A Dimension describes the values that a quantity can take in the context of a design space.

    Abstract type that returns the proper type given a serialized dict.

    """

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            'ContinuousDimension': ContinuousDimension,
            'EnumeratedDimension': EnumeratedDimension
        }[data['type']]


class ContinuousDimension(Serializable['ContinuousDimension'], Dimension):
    """A continuous, real-valued dimension.

    Parameters
    ----------
    descriptor: RealDescriptor
        a descriptor of the single dimension
    lower_bound: float
        inclusive lower bound
    upper_bound: float
        inclusive upper bound

    """

    descriptor = properties.Object(RealDescriptor, 'descriptor')
    lower_bound = properties.Float('lower_bound')
    upper_bound = properties.Float('upper_bound')
    typ = properties.String('type', default='ContinuousDimension', deserializable=False)

    def __init__(self,
                 descriptor: RealDescriptor, *,
                 lower_bound: Optional[float] = None,
                 upper_bound: Optional[float] = None):
        self.descriptor: RealDescriptor = descriptor
        if lower_bound is not None:
            self.lower_bound: float = lower_bound
        else:
            self.lower_bound: float = descriptor.lower_bound
        if upper_bound is not None:
            self.upper_bound: float = upper_bound
        else:
            self.upper_bound: float = descriptor.upper_bound


class EnumeratedDimension(Serializable['EnumeratedDimension'], Dimension):
    """A finite, enumerated dimension.

    Parameters
    ----------
    descriptor: Descriptor
        a descriptor of the single dimension
    values: list[str]
        list of values that can be parsed by the descriptor

    """

    descriptor = properties.Object(Descriptor, 'descriptor')
    values = properties.List(properties.String(), 'list')
    typ = properties.String('type', default='EnumeratedDimension', deserializable=False)

    def __init__(self,
                 descriptor: Descriptor, *,
                 values: List[str]):
        self.descriptor: Descriptor = descriptor
        self.values: List[str] = values



import json

import jwt
import pytest
import unittest

from citrine.exceptions import (
    BadRequest,
    CitrineException,
    Conflict,
    NonRetryableException,
    WorkflowNotReadyException,
    RetryableException)

from datetime import datetime, timedelta
import pytz
import mock
import requests
import requests_mock
from citrine._session import Session
from citrine.exceptions import UnauthorizedRefreshToken, Unauthorized, NotFound
from tests.utils.session import make_fake_cursor_request_function


def refresh_token(expiration: datetime = None) -> dict:
    token = jwt.encode(
        payload={'exp': expiration.timestamp()},
        key='garbage'
    )
    return {'access_token': token}


@pytest.fixture
def session():
    token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))
    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/utils/runtime-config', json=dict())
        m.post('http://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        session = Session(
            refresh_token='12345',
            scheme='http',
            host='citrine-testing.fake'
        )
    # Default behavior is to *not* require a refresh - those tests can clear this out
    # As rule of thumb, we should be using freezegun or similar to never rely on the system clock
    # for these scenarios, but I thought this is light enough to postpone that for the time being
    session.access_token_expiration = datetime.utcnow() + timedelta(minutes=3)

    return session


def test_session_signature(monkeypatch):
    token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))
    with requests_mock.Mocker() as m:
        m.post('ftp://citrine-testing.fake:8080/api/v1/tokens/refresh', json=token_refresh_response)
        m.get('ftp://citrine-testing.fake:8080/api/v1/utils/runtime-config', json=dict())

        assert '1234' == Session(refresh_token='1234',
                                 scheme='ftp',
                                 host='citrine-testing.fake',
                                 port="8080").refresh_token

    # Validate defaults
    with requests_mock.Mocker() as m:
        patched_key = "5678"
        patched_host = "monkeypatch.citrine-testing.fake"
        monkeypatch.setenv("CITRINE_API_KEY", patched_key)
        monkeypatch.setenv("CITRINE_API_HOST", patched_host)
        m.post(f'https://{patched_host}/api/v1/tokens/refresh', json=token_refresh_response)
        m.get(f'https://{patched_host}/api/v1/utils/runtime-config', json=dict())

        assert patched_key == Session().refresh_token
        assert patched_key == Session(patched_key).refresh_token
        monkeypatch.delenv("CITRINE_API_KEY")
        assert Session().refresh_token is None

    monkeypatch.delenv("CITRINE_API_HOST")
    with pytest.raises(ValueError):
        Session()


def test_deprecated_positional():
    token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))
    with requests_mock.Mocker() as m:
        m.post('ftp://citrine-testing.fake:8080/api/v1/tokens/refresh', json=token_refresh_response)
        m.get('ftp://citrine-testing.fake:8080/api/v1/utils/runtime-config', json=dict())

        with pytest.warns(DeprecationWarning):
            assert '1234' == Session('1234', 'ftp', 'citrine-testing.fake', "8080").refresh_token

    with pytest.warns(DeprecationWarning):
        with pytest.raises(ValueError):
            Session('1234', 'ftp', scheme='ftp')


def test_get_refreshes_token(session: Session):
    session.access_token_expiration = datetime.utcnow() - timedelta(minutes=1)
    token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))

    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/utils/runtime-config', json=dict())
        m.post('http://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        m.get('http://citrine-testing.fake/api/v1/foo',
              json={'foo': 'bar'},
              headers={'content-type': "application/json"})

        resp = session.get_resource('/foo')

    assert {'foo': 'bar'} == resp
    assert datetime(2019, 3, 14) == session.access_token_expiration


def test_get_runtime_config_failure(session: Session):
    session.access_token_expiration = datetime.utcnow() - timedelta(minutes=1)
    token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))

    with requests_mock.Mocker() as m:
        m.post('http://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        m.get('http://citrine-testing.fake/api/v1/utils/runtime-config', status_code=400)

        with pytest.raises(CitrineException):
            session.get_resource('/foo')


def test_get_refresh_token_failure(session: Session):
    session.access_token_expiration = datetime.utcnow() - timedelta(minutes=1)

    with requests_mock.Mocker() as m:
        m.post('http://citrine-testing.fake/api/v1/tokens/refresh', status_code=401)

        with pytest.raises(UnauthorizedRefreshToken):
            session.get_resource('/foo')


def test_get_no_refresh(session: Session):
    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/foo', json={'foo': 'bar'}, headers={'content-type': "application/json"})
        resp = session.get_resource('/foo')

    assert {'foo': 'bar'} == resp


def test_get_not_found(session: Session):
    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/foo', status_code=404)
        with pytest.raises(NotFound):
            session.get_resource('/foo')


def test_status_code_409(session: Session):
    with requests_mock.Mocker() as m:
        url = '/foo'
        conflict_message = 'you have a conflict'
        resp_json = {
            'code': 409,
            'message': 'a message',
            'validation_errors': [{'failure_message': conflict_message}]
        }
        m.get('http://citrine-testing.fake/api/v1/foo', status_code=409, json=resp_json)
        with pytest.raises(Conflict) as einfo:
            session.get_resource(url)

        assert getattr(einfo.value, "url", None) == url
        assert conflict_message in str(einfo.value)
        assert "409" in str(einfo.value)


def test_status_code_425(session: Session):
    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/foo', status_code=425)
        with pytest.raises(RetryableException):
            session.get_resource('/foo')
        with pytest.raises(WorkflowNotReadyException):
            session.get_resource('/foo')


def test_status_code_400(session: Session):
    with requests_mock.Mocker() as m:
        resp_json = {
            'code': 400,
            'message': 'a message',
            'validation_errors': [
                {
                    'failure_message': 'you have failed',
                },
            ],
        }
        m.get('http://citrine-testing.fake/api/v1/foo',
              status_code=400,
              json=resp_json
              )
        with pytest.raises(BadRequest) as einfo:
            session.get_resource('/foo')
        assert einfo.value.api_error.validation_errors[0].failure_message \
               == resp_json['validation_errors'][0]['failure_message']


def test_status_code_401(session: Session):
    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/foo', status_code=401)
        with pytest.raises(NonRetryableException):
            session.get_resource('/foo')
        with pytest.raises(Unauthorized):
            session.get_resource('/foo')


def test_status_code_404(session: Session):
    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/foo', status_code=404)
        with pytest.raises(NonRetryableException):
            session.get_resource('/foo')


def test_connection_error(session: Session):
    data = {'stuff': 'not_used'}

    # Simulate a request using a stale session that raises
    # a ConnectionError then works on the second call.
    with requests_mock.Mocker() as m:
        m.register_uri('GET',
                       'http://citrine-testing.fake/api/v1/foo',
                       [{'exc': requests.exceptions.ConnectionError},
                        {'json': data, 'status_code': 200, 'headers': {'content-type': "application/json"}}])

        resp = session.get_resource('/foo')
        assert resp == data


def test_post_refreshes_token_when_denied(session: Session):
    token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))

    with requests_mock.Mocker() as m:
        m.get('http://citrine-testing.fake/api/v1/utils/runtime-config', json=dict())
        m.post('http://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        m.register_uri('POST', 'http://citrine-testing.fake/api/v1/foo', [
            {'status_code': 401, 'json': {'reason': 'invalid-token'}},
            {'json': {'foo': 'bar'}, 'headers': {'content-type': "application/json"}}
        ])

        resp = session.post_resource('/foo', json={'data': 'hi'})

    assert {'foo': 'bar'} == resp
    assert datetime(2019, 3, 14) == session.access_token_expiration


# this test exists to provide 100% coverage for the legacy 401 status on Unauthorized responses
def test_delete_unauthorized_without_json_legacy(session: Session):
    with requests_mock.Mocker() as m:
        m.delete('http://citrine-testing.fake/api/v1/bar/something', status_code=401)

        with pytest.raises(Unauthorized):
            session.delete_resource('/bar/something')


def test_delete_unauthorized_with_str_json_legacy(session: Session):
    with requests_mock.Mocker() as m:
        m.delete(
            'http://citrine-testing.fake/api/v1/bar/something',
            status_code=401,
            json='an error string'
        )

        with pytest.raises(Unauthorized):
            session.delete_resource('/bar/something')


def test_delete_unauthorized_without_json(session: Session):
    with requests_mock.Mocker() as m:
        m.delete('http://citrine-testing.fake/api/v1/bar/something', status_code=403)

        with pytest.raises(Unauthorized):
            session.delete_resource('/bar/something')


def test_failed_put_with_stacktrace(session: Session):
    with mock.patch("time.sleep", return_value=None):
        with requests_mock.Mocker() as m:
            m.put(
                'http://citrine-testing.fake/api/v1/bad-endpoint',
                status_code=500,
                json={'debug_stacktrace': 'blew up!'}
            )

            with pytest.raises(Exception) as e:
                session.put_resource('/bad-endpoint', json={})

        assert '{"debug_stacktrace": "blew up!"}' == str(e.value)


def test_cursor_paged_resource():
    full_result_set = list(range(26))

    fake_request = make_fake_cursor_request_function(full_result_set)

    # varying page size should not affect final result
    assert list(Session.cursor_paged_resource(fake_request, 'foo', forward=True, per_page=10)) == full_result_set
    assert list(Session.cursor_paged_resource(fake_request, 'foo', forward=True, per_page=26)) == full_result_set
    assert list(Session.cursor_paged_resource(fake_request, 'foo', forward=True, per_page=40)) == full_result_set


def test_bad_json_response(session: Session):
    with requests_mock.Mocker() as m:
        m.delete('http://citrine-testing.fake/api/v1/bar/something',
                 status_code=200,
                 headers={'content-type': "application/json"})
        response_json = session.delete_resource('/bar/something')
        assert response_json == {}


def test_good_json_response(session: Session):
    with requests_mock.Mocker() as m:
        json_to_validate = {"bar": "something"}
        m.put('http://citrine-testing.fake/api/v1/bar/something',
              status_code=200,
              json=json_to_validate,
              headers={'content-type': "application/json"})
        response_json = session.put_resource('bar/something', {"ignored": "true"})
        assert response_json == json_to_validate


def test_patch(session: Session):
    with requests_mock.Mocker() as m:
        json_to_validate = {"bar": "something"}
        m.patch('http://citrine-testing.fake/api/v1/bar/something',
                status_code=200,
                json=json_to_validate,
                headers={'content-type': "application/json"})
        response_json = session.patch_resource('bar/something', {"ignored": "true"})
        assert response_json == json_to_validate



from typing import Optional, Type
from uuid import UUID

from deprecation import deprecated

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._session import Session
from citrine.resources.report import ReportResource


__all__ = ['Predictor']


class Predictor(PolymorphicSerializable['Predictor'], AsynchronousObject):
    """Module that describes the ability to compute/predict properties of materials.

    Abstract type that returns the proper type given a serialized dict. Subtype
    based on the 'type' value of the passed in dict.

    """

    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    """:Optional[UUID]: Citrine Platform unique identifier"""
    name = properties.String('data.name')
    description = properties.Optional(properties.String(), 'data.description')

    _response_key = None
    _project_id: Optional[UUID] = None
    _session: Optional[Session] = None
    _in_progress_statuses = ["VALIDATING", "CREATED"]
    _succeeded_statuses = ["READY"]
    _failed_statuses = ["INVALID", "ERROR"]

    @property
    @deprecated(deprecated_in="1.31.0", removed_in="2.0.0",
                details="Please use `isinstance` or `issubclass` instead.")
    def module_type(self):
        """The type of module."""
        return "PREDICTOR"

    @module_type.setter
    @deprecated(deprecated_in="1.31.0", removed_in="2.0.0")
    def module_type(self, value):
        pass

    @property
    def report(self):
        """Fetch the predictor report."""
        if self.uid is None or self._session is None or self._project_id is None \
                or getattr(self, "version", None) is None:
            msg = "Cannot get the report for a predictor that wasn't read from the platform"
            raise ValueError(msg)
        report_resource = ReportResource(self._project_id, self._session)
        return report_resource.get(predictor_id=self.uid, predictor_version=self.version)

    @staticmethod
    def wrap_instance(predictor_data: dict) -> dict:
        """Insert a serialized embedded predictor into a module envelope.

        This facilitates deserialization.
        """
        return {
            "data": {
                "name": predictor_data.get("name", ""),
                "description": predictor_data.get("description", ""),
                "instance": predictor_data
            }
        }

    @classmethod
    def get_type(cls, data) -> Type['Predictor']:
        """Return the subtype."""
        from .simple_ml_predictor import SimpleMLPredictor
        from .graph_predictor import GraphPredictor
        from .expression_predictor import ExpressionPredictor
        from .molecular_structure_featurizer import MolecularStructureFeaturizer
        from .ingredients_to_formulation_predictor import IngredientsToFormulationPredictor
        from .label_fractions_predictor import LabelFractionsPredictor
        from .simple_mixture_predictor import SimpleMixturePredictor
        from .ingredient_fractions_predictor import IngredientFractionsPredictor
        from .auto_ml_predictor import AutoMLPredictor
        from .mean_property_predictor import MeanPropertyPredictor
        from .chemical_formula_featurizer import ChemicalFormulaFeaturizer
        type_dict = {
            "Simple": SimpleMLPredictor,
            "Graph": GraphPredictor,
            "AnalyticExpression": ExpressionPredictor,
            "MoleculeFeaturizer": MolecularStructureFeaturizer,
            "IngredientsToSimpleMixture": IngredientsToFormulationPredictor,
            "MeanProperty": MeanPropertyPredictor,
            "LabelFractions": LabelFractionsPredictor,
            "SimpleMixture": SimpleMixturePredictor,
            "IngredientFractions": IngredientFractionsPredictor,
            "ChemicalFormulaFeaturizer": ChemicalFormulaFeaturizer,
            "AutoML": AutoMLPredictor,
        }
        typ = type_dict.get(data['data']['instance']['type'])

        if typ is not None:
            return typ
        else:
            raise ValueError(
                '{} is not a valid predictor type. '
                'Must be in {}.'.format(data['data']['instance']['type'], type_dict.keys())
            )



"""Resources that represent both individual and collections of teams."""
from typing import Optional, Union, List
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.resources import ProjectCollection
from citrine.resources.user import User, UserCollection

WRITE = "WRITE"
READ = "READ"
SHARE = "SHARE"
TEAM_ACTIONS = Union[WRITE, READ, SHARE]


class TeamMember:
    """A Member of a Team."""

    def __init__(self,
                 *,
                 user: User,
                 team: 'Team',  # noqa: F821
                 actions: List[TEAM_ACTIONS]):
        self.user = user
        self.team: 'Team' = team  # noqa: F821
        self.actions: List[TEAM_ACTIONS] = actions

    def __str__(self):
        return '<TeamMember {!r} can {!s} in {!r}>' \
            .format(self.user.screen_name, self.actions, self.team.name)


class TeamResourceIDs:
    """
    A Citrine Team Resource IDs class.

    This class is used to list the unique IDs for specific resource
    types published in a single team and therefore available to be pulled
    in by all projects.

    Parameters
    ----------
    team_id: str or uuid
        ID of the team.
    resource_type: str
        The resource type to list, one of DATASET/MODULE/TABLE/TABLE_DEFINITION

    """

    _api_version = "v3"

    def __init__(self,
                 session: Session,
                 team_id: Union[str, UUID],
                 resource_type: str) -> None:
        self.session = session
        self.team_id = team_id
        self.resource_type = resource_type

    def _path(self) -> str:
        return format_escaped_url(f'/teams/{self.team_id}')

    def _list_ids(self, action: str) -> List[str]:
        query_params = {"domain": self._path(), "action": action}
        return self.session.get_resource(f"/{self.resource_type}/authorized-ids",
                                         params=query_params,
                                         version=self._api_version)['ids']

    def list_readable(self):
        """
        List IDs of the published resources with READ access.

        Returns
        -------
        List[str]
            Returns a list of string UUIDs for each resource

        """
        return self._list_ids(action=READ)

    def list_writeable(self):
        """
        List IDs of the published resources with WRITE access.

        Returns
        -------
        List[str]
            Returns a list of string UUIDs for each resource

        """
        return self._list_ids(action=WRITE)

    def list_shareable(self):
        """
        List IDs of the published resources with SHARE access.

        Returns
        -------
        List[str]
            Returns a list of string UUIDs for each resource

        """
        return self._list_ids(action=SHARE)


class Team(Resource['Team']):
    """
    A Citrine Team.

    A Team is a collection of projects and resources.

    Parameters
    ----------
    name: str
        Name of the team.
    description: str
        Long-form description of the team.
    session: Session, optional
        The Citrine session used to connect to the database.

    Attributes
    ----------
    uid: UUID
        Unique uuid4 identifier of this team.
    created_at: int
        Time the team was created, in seconds since epoch.
    name: str
        Name of the Team
    description: str
        Description of the Team

    """

    _response_key = 'team'
    _resource_type = ResourceTypeEnum.TEAM
    _api_version = "v3"

    name = properties.String('name')
    description = properties.Optional(properties.String(), 'description')
    uid = properties.Optional(properties.UUID(), 'id')
    created_at = properties.Optional(properties.Datetime(), 'created_at')

    def __init__(self,
                 name: str,
                 *,
                 description: str = "",
                 session: Optional[Session] = None):
        self.name: str = name
        self.description: str = description
        self.session: Session = session

    def __str__(self):
        return '<Team {!r}>'.format(self.name)

    def _path(self):
        return format_escaped_url('/teams/{team_id}', team_id=self.uid)

    def list_members(self) -> List[TeamMember]:
        """
        List all of the members in the current team.

        Requires admin privileges.

        Returns
        -------
        List[TeamMember]
            The members of the current team

        """
        members = self.session.get_resource(self._path() + "/users",
                                            version=self._api_version)["users"]
        return [TeamMember(user=User.build(m), team=self, actions=m["actions"]) for m in members]

    def get_member(self, user_id: Union[str, UUID, User]) -> TeamMember:
        """
        Get a particular member in the current team.

        May require admin privileges depending on which user is being requested.

        Parameters
        ----------
        user_id: str or uuid
            The id of the user to remove from the team

        Returns
        -------
        TeamMember
            The requested team member

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        path = self._path() + format_escaped_url('/users/{user_id}', user_id=user_id)
        member = self.session.get_resource(path=path, version=self._api_version)["user"]
        return TeamMember(user=User.build(member), team=self, actions=member["actions"])

    def me(self) -> TeamMember:
        """
        Get the member for the current user.

        Returns
        -------
        TeamMember
            The TeamMember object representing the current user

        """
        me = UserCollection(self.session).me()
        return self.get_member(me)

    def remove_user(self, user_id: Union[str, UUID, User]) -> bool:
        """
        Remove a User from a Team.

        Requires admin privileges.

        Parameters
        ----------
        user_id: User, str or uuid
            The id of the user to remove from the team

        Returns
        -------
        bool
            Returns ``True`` if user successfully removed

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        self.session.checked_post(self._path() + "/users/batch-remove",
                                  json={"ids": [str(user_id)]}, version=self._api_version)
        return True  # note: only get here if checked_post doesn't raise error

    def add_user(self,
                 user_id: Union[str, UUID, User],
                 *,
                 actions: Optional[List[TEAM_ACTIONS]] = None) -> bool:
        """
        Add a User to a Team.

        Requires admin privileges.

        If no actions are specified, adds User with ``READ`` action to the Team.

        Use the `update_user_action` method to change a User's actions.

        Parameters
        ----------
        user_id: User, str or uuid
            The id of the user to add to the team
        actions: list of TEAM_ACTIONS
            The actions to give the new user in this team
            The options are: WRITE, READ, SHARE
            defaults to READ

        Returns
        -------
        bool
            Returns ``True`` if user successfully added

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        if actions is None:
            actions = [READ]
        return self.update_user_action(user_id, actions=actions)

    def update_user_action(self,
                           user_id: Union[str, UUID, User],
                           *,
                           actions: List[TEAM_ACTIONS]) -> bool:
        """
        Overwrites a User's action permissions in the Team.

        Requires admin privileges.

        Parameters
        ----------
        user_id: User, str or uuid
            The id of the user to add to the team
        actions: list of TEAM_ACTIONS
            The actions to give the new user in this team
            The options are: WRITE, READ, SHARE

        Returns
        -------
        bool
            Returns ``True`` if user successfully added

        """
        if isinstance(user_id, User):
            user_id = user_id.uid
        self.session.checked_put(self._path() + "/users", version=self._api_version,
                                 json={'id': str(user_id), "actions": actions})
        return True

    def share(self,
              *,
              resource: Resource,
              target_team_id: Union[str, UUID, "Team"]) -> bool:
        """
        Share a resource with another team.

        Requires SHARE action.

        Parameters
        ----------
        resource: Resource
            The resource owned by this team, which will be shared
        target_team_id: Union[str, UUID, Team]
            The id of the team with which to share the resource

        Returns
        -------
        bool
            Returns ``True`` if resource successfully shared

        """
        if isinstance(target_team_id, Team):
            target_team_id = target_team_id.uid
        resource_access = resource.access_control_dict()
        payload = {
            "resource_type": resource_access["type"],
            "resource_id": resource_access["id"],
            "target_team_id": str(target_team_id)
        }
        self.session.checked_post(self._path() + "/shared-resources",
                                  version=self._api_version, json=payload)
        return True

    def un_share(self, *, resource: Resource, target_team_id: Union[str, UUID, "Team"]) -> bool:
        """
        Revoke the share of a particular resource to a secondary team.

        Requires SHARE action.

        Parameters
        ----------
        resource: Resource
            The resource owned by this team, which will be un-shared
        target_team_id: Union[str, UUID, Team]
            The id of the team which should not have access to the resource

        Returns
        -------
        bool
            Returns ``True`` if resource successfully un-shared

        """
        if isinstance(target_team_id, Team):
            target_team_id = target_team_id.uid
        resource_type = resource.access_control_dict()["type"]
        resource_id = resource.access_control_dict()["id"]
        self.session.checked_delete(
            self._path() + f"/shared-resources/{resource_type}/{resource_id}",
            version=self._api_version,
            json={"target_team_id": str(target_team_id)}
        )
        return True

    @property
    def projects(self) -> ProjectCollection:
        """Return a resource representing all visible projects in this team."""
        return ProjectCollection(self.session, team_id=self.uid)

    @property
    def dataset_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published dataset IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.DATASET.value)

    @property
    def module_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published module IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.MODULE.value)

    @property
    def table_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published table IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.TABLE.value)

    @property
    def table_definition_ids(self) -> TeamResourceIDs:
        """Return a TeamResourceIDs instance for listing published table definition IDs."""
        return TeamResourceIDs(session=self.session,
                               team_id=self.uid,
                               resource_type=ResourceTypeEnum.TABLE_DEFINITION.value)


class TeamCollection(Collection[Team]):
    """
    Represents the collection of all teams as well as the resources belonging to it.

    Parameters
    ----------
    session: Session, optional
        The Citrine session used to connect to the database.

    """

    _path_template = '/teams'
    _individual_key = 'team'
    _collection_key = 'teams'
    _resource = Team
    _api_version = "v3"

    def __init__(self, session: Session):
        self.session = session

    def build(self, data) -> Team:
        """
        Build an individual team from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the team.

        Return
        -------
        Team
            The team created from data.

        """
        team = Team.build(data)
        team.session = self.session
        return team

    def update(self, team: Team) -> Team:
        """
        Update a particular team.

        You can only update the name and/or description.
        Parameters
        ----------
        team: Team
            The Team Resource to be updated

        """
        url = self._get_path(team.uid)
        updated = self.session.patch_resource(url, team.dump(), version=self._api_version)
        data = updated[self._individual_key]
        return self.build(data)

    def register(self, name: str, *, description: str = "") -> Team:
        """
        Create and upload new team.

        Automatically grants user READ, WRITE, and SHARE access to this team.

        Parameters
        ----------
        name: str
            Name of the team to be created.
        description: str
            Long-form description of the team to be created.

        """
        team = super().register(Team(name, description=description))
        user_id = UserCollection(self.session).me().uid
        team.update_user_action(user_id=user_id, actions=[READ, WRITE, SHARE])
        return team



from typing import List, Optional, Union
from uuid import UUID

from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.predictors import Predictor

__all__ = ['GraphPredictor']


class GraphPredictor(VersionedEngineResource['GraphPredictor'], Predictor):
    """A predictor interface that stitches other predictors together.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    predictors: List[Union[UUID, Predictor]]
        the list of predictors to use in the graph, either UUIDs or serialized predictors
    training_data: Optional[List[DataSource]]
        Optional sources of training data shared by all predictors in the graph.
        Training data provided by this graph predictor does not need to be specified as part of the
        configuration of sub-predictors. Shared training data and any training data specified
        by a sub-predictor will be combined into a flattened list and de-duplicated
        by uid and identifiers. De-duplication is performed if a uid or identifier is shared
        between two or more rows. The content of a de-duplicated row will contain the union of
        data across all rows that share the same uid or at least 1 identifier.

    """

    predictors = _properties.List(_properties.Union(
        [_properties.UUID, _properties.Object(Predictor)]), 'data.instance.predictors')
    # the default seems to be defined in instances, not the class itself
    # this is tested in test_graph_default_training_data
    training_data = _properties.List(_properties.Object(DataSource),
                                     'data.instance.training_data', default=[])

    typ = _properties.String('data.instance.type', default='Graph', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 predictors: List[Union[UUID, Predictor]],
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description
        self.predictors: List[Union[UUID, Predictor]] = predictors
        self.training_data: List[DataSource] = training_data or []

    def _post_dump(self, data: dict) -> dict:
        data = super()._post_dump(data)
        for i, predictor in enumerate(data['instance']['predictors']):
            if isinstance(predictor, dict):
                # embedded predictors are not modules, so only serialize their config
                data['instance']['predictors'][i] = predictor['instance']
        return data

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        for i, predictor_data in enumerate(data['data']['instance']['predictors']):
            if isinstance(predictor_data, dict):
                data['data']['instance']['predictors'][i] = Predictor.wrap_instance(predictor_data)
        return data

    def __str__(self):
        return '<GraphPredictor {!r}>'.format(self.name)



"""A collection of FileLink objects."""
import mimetypes
import os
from pathlib import Path
from enum import Enum
from logging import getLogger
from typing import Iterable, Optional, Tuple, Union, List, Dict
from urllib.parse import urlparse, quote
from uuid import UUID

import requests
from boto3 import client as boto3_client
from boto3.session import Config
from botocore.exceptions import ClientError
from citrine._rest.collection import Collection
from citrine._rest.resource import Resource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String, Object, Integer
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine._utils.functions import rewrite_s3_links_locally
from citrine._utils.functions import write_file_locally, format_escaped_url

from citrine.jobs.job import JobSubmissionResponse, _poll_for_job_completion
from citrine.resources.response import Response
from gemd.entity.bounds.base_bounds import BaseBounds
from gemd.entity.file_link import FileLink as GEMDFileLink
from gemd.enumeration.base_enumeration import BaseEnumeration

logger = getLogger(__name__)


class SearchFileFilterTypeEnum(BaseEnumeration):
    """
    The type of the filter used to search for files.

    * SEARCH_BY_NAME:
        Search a file by name in a specific dataset,
        returns by default the last version or a specific one
    * SEARCH_BY_VERSION_ID:
        Search by a specific file version id
    * SEARCH_BY_DATASET_FILE_ID:
        Search either the last version or a specific version number for a specific dataset file id

    """

    NAME_SEARCH = "search_by_name"
    VERSION_ID_SEARCH = "search_by_version_id"
    DATASET_FILE_ID_SEARCH = "search_by_dataset_file_id"


class _Uploader:
    """Holds the many parameters that are generated and used during file upload."""

    def __init__(self):
        self.bucket = ''
        self.object_key = ''
        self.upload_id = ''
        self.region_name = ''
        self.aws_access_key_id = ''
        self.aws_secret_access_key = ''
        self.aws_session_token = ''
        self.s3_version = ''
        self.s3_endpoint_url = None
        self.s3_use_ssl = True
        self.s3_addressing_style = 'auto'


class FileProcessingType(Enum):
    """The supported File Processing Types."""

    VALIDATE_CSV = "VALIDATE_CSV"


class FileProcessingData:
    """The base class of all File Processing related data implementations."""

    pass


class CsvColumnInfo(Serializable):
    """The info for a CSV Column, contains the name, recommended and exact bounds."""

    name = String('name')
    """:str: name of the column"""
    bounds = Object(BaseBounds, 'bounds')
    """:BaseBounds: recommended bounds of the column (might include some padding)"""
    exact_range_bounds = Object(BaseBounds, 'exact_range_bounds')
    """:BaseBounds: exact bounds of the column"""

    def __init__(self, name: String, bounds: BaseBounds,
                 exact_range_bounds: BaseBounds):  # pragma: no cover
        self.name = name
        self.bounds = bounds
        self.exact_range_bounds = exact_range_bounds


class CsvValidationData(FileProcessingData, Serializable):
    """The resulting data from the processed CSV file."""

    columns = PropertyOptional(PropertyList(Object(CsvColumnInfo)), 'columns',
                               override=True)
    """:Optional[List[CsvColumnInfo]]: all of the columns in the CSV"""
    record_count = Integer('record_count')
    """:int: the number of rows in the CSV"""

    def __init__(self, columns: List[CsvColumnInfo],
                 record_count: int):  # pragma: no cover
        self.columns = columns
        self.record_count = record_count


class FileProcessingResult:
    """
    The results of a successful file processing operation.

    The type of the actual data depends on the specific processing type.
    """

    def __init__(self, processing_type: FileProcessingType, data: Union[Dict,
                                                                        FileProcessingData]):
        self.processing_type = processing_type
        self.data = data


class FileLink(Resource['FileLink'], GEMDFileLink):
    """
    Resource that stores the name and url of an external file.

    Parameters
    ----------
    filename: str
        The name of the file.
    url: str
        URL that can be used to access the file.

    """

    filename = String('filename', override=True)
    url = String('url', override=True)
    typ = String('type')

    def __init__(self, filename: str, url: str):
        GEMDFileLink.__init__(self, filename, url)
        self.typ = GEMDFileLink.typ

    @property
    def name(self):
        """Attribute name is an alias for filename."""
        return self.filename

    def __str__(self):
        return '<File link {!r}>'.format(self.filename)

    def as_dict(self) -> dict:
        """Dump to a dictionary (useful for interoperability with gemd)."""
        return self.dump()


class FileCollection(Collection[FileLink]):
    """Represents the collection of all file links associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/files'
    _collection_key = 'files'
    _resource = FileLink

    def __init__(self, project_id: UUID, dataset_id: UUID, session: Session):
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.session = session

    def _get_path(self,
                  uid: Optional[Union[UUID, str]] = None,
                  ignore_dataset: Optional[bool] = False,
                  *,
                  version: Union[str, UUID] = None,
                  action: str = None) -> str:
        """Build the path for taking an action with a particular file version."""
        base = urlparse(super()._get_path(uid=uid, ignore_dataset=ignore_dataset))
        new_path = base.path.split('/')
        if version is not None:
            new_path.extend([quote(x) for x in ('versions', str(version))])
        if action is not None:
            new_path.append(quote(action))

        return base._replace(path='/'.join(new_path)).geturl()

    @staticmethod
    def _get_ids_from_url(url: str) -> Tuple[Optional[UUID], Optional[UUID]]:
        """Attempt to extract file_id and version_id from a URL."""
        parsed = urlparse(url)
        if len(parsed.query) > 0 or len(parsed.fragment) > 0:
            # Illegal modifiers
            return None, None
        split_path = urlparse(url).path.split('/')
        if len(split_path) >= 4 and split_path[-4] == 'files' and split_path[-2] == 'versions':
            file_id = split_path[-3]
            version_id = split_path[-1]
        elif len(split_path) >= 2 and split_path[-2] == 'files':
            file_id = split_path[-1]
            version_id = None
        else:
            file_id, version_id = None, None

        if file_id is not None:
            try:
                file_id = UUID(file_id)
                if version_id is not None:
                    version_id = UUID(version_id)
            except ValueError:
                return None, None
        return file_id, version_id

    def _get_path_from_file_link(self, file_link: FileLink,
                                 *,
                                 action: str = None) -> str:
        """Build the platform path for taking an action with a particular file link."""
        # Use this sessions project/dataset credentials and the URL's file / version
        file_id, version_id = self._get_ids_from_url(file_link.url)
        if file_id is None:
            raise ValueError("FileLink did not contain a Citrine platform file URL.")
        return self._get_path(uid=file_id, version=version_id, action=action)

    def build(self, data: dict) -> FileLink:
        """Build an instance of FileLink."""
        return FileLink.build(data)

    def _fetch_page(self,
                    page: Optional[int] = None,
                    per_page: Optional[int] = None) -> Tuple[Iterable[FileLink], str]:
        """
        List all visible files in the collection.

        Parameters
        ---------
        page: int, optional
            The "page" number of results to list. Default is the first page, which is 1.
        per_page: int, optional
            Max number of results to return for each call. Default is 20.

        Returns
        -------
        Iterable[FileLink]
            FileLink objects in this collection.
        str
            The next uri if one is available, empty string otherwise

        """
        path = self._get_path()
        params = {}
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self.session.get_resource(path=path, params=params)
        collection = response[self._collection_key]
        return collection, ""

    def _build_collection_elements(self, collection):
        for file in collection:
            yield self.build(self._as_dict_from_resource(file))

    def _as_dict_from_resource(self, file: dict):
        """
        Convert a file link resource downloaded from the API into a FileLink dictionary.

        This is necessary because the database resource contains additional information that is
        not in the FileLink object, such as file size and the id of the user who uploaded the file.

        Parameters
        ---------
        file: dict
            A JSON dictionary corresponding to the file link as it is saved in the database.

        Returns
        -------
        dict
            A dictionary that can be built into a FileLink object.

        """
        # FIXME  While the 'id' field is supposed to be the file ID, it contains the version
        #  for some reason.  Needs to be fixed on back end.  PLA-9482
        filename = file['filename']
        file_id = file['id']
        version_id = file['version']

        file_dict = {
            'url': self._get_path(uid=file_id, version=version_id),
            'filename': filename,
            'type': GEMDFileLink.typ
        }
        return file_dict

    def get(self,
            uid: Union[UUID, str],
            *,
            version: Optional[Union[UUID, str, int]] = None) -> FileLink:
        """
        Get an element of the collection by its id.

        Parameters
        ----------
        uid: Union[UUID, str]
            A representation of the FileLink (Citrine id or file name)
        version: Optional[UUID, str, int]
            The version, as a UUID or str(UUID) of the version_id or an int or
            str(int) of the version number.  If None, returns the file with the
            highest version number (most recent).

        Returns
        -------
        ResourceType
            An object with specified scope and uid

        """
        if not isinstance(uid, (str, UUID)):
            raise TypeError(f"File Link can only be resolved from str or UUID."
                            f"Instead got {type(uid)} {uid}.")
        if version is not None and not isinstance(version, (str, UUID, int)):
            raise TypeError(f"Version can only be resolved from str, int or UUID."
                            f"Instead got {type(uid)} {uid}.")

        if isinstance(uid, str):
            try:  # Check if the uid string is actually a UUID
                uid = UUID(uid)
            except ValueError:
                pass

        if isinstance(version, str):
            try:  # Check if the version string is actually a UUID
                version = UUID(version)
            except ValueError:
                try:  # Check if the version string is actually an int / version number
                    version = int(version)
                except ValueError:
                    raise ValueError(
                        f"Version {version} could not be converted to either an int or a UUID"
                    )

        if isinstance(uid, str):
            # Assume it's the filename on platform;
            if version is None or isinstance(version, int):
                file = self._search_by_file_name(dset_id=self.dataset_id,
                                                 file_name=uid,
                                                 file_version_number=version)
            else:  # We did our type checks earlier; version is an UUID
                file = self._search_by_file_version_id(file_version_id=version)
        else:  # We did our type checks earlier; uid is a UUID
            if isinstance(version, UUID):
                file = self._search_by_file_version_id(file_version_id=version)
            else:  # We did our type checks earlier; version is an int or None
                file = self._search_by_dataset_file_id(dataset_file_id=uid,
                                                       dset_id=self.dataset_id,
                                                       file_version_number=version)

        return file

    def upload(self, *, file_path: Union[str, Path], dest_name: str = None) -> FileLink:
        """
        Uploads a file to the dataset.

        Parameters
        ----------
        file_path: str, Path
            The path to the file on the local computer.
        dest_name: str, optional
            The name the file will have after being uploaded. If unspecified, the local name of
            the file will be used. That is, the file at "/Users/me/diagram.pdf" will be uploaded
            with the name "diagram.pdf". File names **must be unique** within a dataset. If a file
            is uploaded with the same `dest_name` as an existing file it will be considered
            a new version of the existing file.

        Returns
        -------
        FileLink
            The filename and url of the uploaded object.

        """
        file_path = Path(file_path).expanduser()
        if not file_path.is_file():
            raise ValueError(f"{file_path} is not a file.")

        if not dest_name:
            # Use the file name as a default dest_name
            dest_name = file_path.name

        uploader = self._make_upload_request(file_path, dest_name)
        uploader = self._upload_file(file_path, uploader)
        return self._complete_upload(dest_name, uploader)

    def _make_upload_request(self, file_path: Path, dest_name: str):
        """
        Make a request to the backend to upload a file. Uses mimetypes.guess_type.

        Parameters
        ----------
        file_path: Path
            The path to the file on the local computer.
        dest_name: str
            The name the file will have after being uploaded.

        Returns
        -------
        _Uploader
            Holds the parameters returned by the upload request, for later use.
            These must include region_name, aws_access_key_id, aws_secret_access_key,
            aws_session_token, bucket, object_key, & upload_id.

        """
        path = self._get_path() + "/uploads"
        mime_type = self._mime_type(file_path)
        file_size = file_path.stat().st_size
        assert isinstance(file_size, int)
        upload_json = {
            'files': [
                {
                    'file_name': dest_name,
                    'mime_type': mime_type,
                    'size': file_size
                }
            ]
        }
        # POST request creates space in S3 for the file and returns AWS-related information
        # (such as temporary credentials) that allow the file to be uploaded.
        upload_request = self.session.post_resource(path=path, json=upload_json)
        uploader = _Uploader()

        # Extract all relevant information from the upload request
        try:

            uploader.region_name = upload_request['s3_region']
            uploader.aws_access_key_id = upload_request['temporary_credentials']['access_key_id']
            uploader.aws_secret_access_key = \
                upload_request['temporary_credentials']['secret_access_key']
            uploader.aws_session_token = upload_request['temporary_credentials']['session_token']
            uploader.bucket = upload_request['s3_bucket']
            uploader.object_key = upload_request['uploads'][0]['s3_key']
            uploader.upload_id = upload_request['uploads'][0]['upload_id']
            uploader.s3_endpoint_url = self.session.s3_endpoint_url
            uploader.s3_use_ssl = self.session.s3_use_ssl
            uploader.s3_addressing_style = self.session.s3_addressing_style

        except KeyError:
            raise RuntimeError("Upload initiation response is missing some fields: "
                               "{}".format(upload_request))
        return uploader

    def _search_by_file_name(self,
                             file_name: str,
                             dset_id: UUID,
                             file_version_number: Optional[int] = None
                             ) -> Optional[FileLink]:
        """
        Make a request to the backend to search a file by name.

        Note that you can specify a version number, in case you don't, it will
        return the last version by default.

        Parameters
        ----------
        file_name: str
            The name of the file.
        dset_id: UUID
            UUID that represents a dataset.
        file_version_number: Optional[int]
            As optional, you can send a specific version number.

        Returns
        -------
        FileLink
            All the data needed for a file.

        """
        path = self._get_path() + "/search"

        search_json = {
            'fileSearchFilter':
                {
                    'type': SearchFileFilterTypeEnum.NAME_SEARCH.value,
                    'datasetId': str(dset_id),
                    'fileName': file_name,
                    'fileVersionNumber': file_version_number
                }
        }

        data = self.session.post_resource(path=path, json=search_json)

        return self.build(self._as_dict_from_resource(data['files'][0]))

    def _search_by_file_version_id(self,
                                   file_version_id: UUID
                                   ) -> Optional[FileLink]:
        """
        Make a request to the backend to search a file by file version id.

        Parameters
        ----------
        file_version_id: UUID
            UUID that represents a file version id.

        Returns
        -------
        FileLink
            All the data needed for a file.

        """
        path = self._get_path() + "/search"

        search_json = {
            'fileSearchFilter': {
                'type': SearchFileFilterTypeEnum.VERSION_ID_SEARCH.value,
                'fileVersionUuid': str(file_version_id)
            }
        }

        data = self.session.post_resource(path=path, json=search_json)

        return self.build(self._as_dict_from_resource(data['files'][0]))

    def _search_by_dataset_file_id(self,
                                   dataset_file_id: UUID,
                                   dset_id: UUID,
                                   file_version_number: Optional[int] = None
                                   ) -> Optional[FileLink]:
        """
        Make a request to the backend to search a file by dataset file id.

        Note that you can specify a version number, in case you don't, it will
        return the last version by default.

        Parameters
        ----------
        dataset_file_id: UUID
            UUID that represents a dataset file id.
        dset_id: UUID
            UUID that represents a dataset.
        file_version_number: Optional[int]
            As optional, you can send a specific version number

        Returns
        -------
        FileLink
            All the data needed for a file.

        """
        path = self._get_path() + "/search"

        search_json = {
            'fileSearchFilter': {
                'type': SearchFileFilterTypeEnum.DATASET_FILE_ID_SEARCH.value,
                'datasetId': str(dset_id),
                'datasetFileId': str(dataset_file_id),
                'fileVersionNumber': file_version_number
            }
        }

        data = self.session.post_resource(path=path, json=search_json)

        return self.build(self._as_dict_from_resource(data['files'][0]))

    @staticmethod
    def _mime_type(file_path: Path):
        # This string coercion is for supporting pathlib.Path objects in python 3.6
        mime_type = mimetypes.guess_type(str(file_path))[0]
        if mime_type is None:
            mime_type = "application/octet-stream"
        return mime_type

    @staticmethod
    def _upload_file(file_path: Path, uploader: _Uploader):
        """
        Upload a file to S3.

        Parameters
        ----------
        file_path: Path
            The path to the file on the local computer.
        uploader: _Uploader
            Holds the parameters returned by the upload request.

        Returns
        -------
        _Uploader
            The input uploader object with its s3_version field now populated.

        """
        additional_s3_opts = {
            'use_ssl': uploader.s3_use_ssl,
            'config': Config(s3={'addressing_style': uploader.s3_addressing_style})
        }

        if uploader.s3_endpoint_url is not None:
            additional_s3_opts['endpoint_url'] = uploader.s3_endpoint_url

        s3_client = boto3_client('s3',
                                 region_name=uploader.region_name,
                                 aws_access_key_id=uploader.aws_access_key_id,
                                 aws_secret_access_key=uploader.aws_secret_access_key,
                                 aws_session_token=uploader.aws_session_token,
                                 **additional_s3_opts)
        with file_path.open(mode='rb') as f:
            try:
                # NOTE: This is only using the simple PUT logic, not the more sophisticated
                # multipart upload approach that is also available (providing parallel
                # uploads, etc).
                upload_response = s3_client.put_object(
                    Bucket=uploader.bucket,
                    Key=uploader.object_key,
                    Body=f,
                    Metadata={"X-Citrine-Upload-Id": uploader.upload_id})
            except ClientError as e:
                raise RuntimeError(f"Upload of file {file_path} failed with the following "
                                   f"exception: {e}")
        uploader.s3_version = upload_response['VersionId']
        return uploader

    def _complete_upload(self, dest_name: str, uploader: _Uploader):
        """
        Indicate that the upload has finished and determine the file URL.

        Parameters
        ----------
        dest_name: str
            The name the file will have after being uploaded.
        uploader: _Uploader
            Holds the parameters returned by the upload request and the upload response.

        Returns
        -------
        FileLink
            The filename and url of the uploaded object.

        """
        path = self._get_path() + format_escaped_url("/uploads/{}/complete", uploader.upload_id)
        complete_response = self.session.put_resource(path=path,
                                                      json={'s3_version': uploader.s3_version})

        try:
            file_id = complete_response['file_info']['file_id']
            version_id = complete_response['file_info']['version']
        except KeyError:
            raise RuntimeError("Upload completion response is missing some "
                               "fields: {}".format(complete_response))

        url = self._get_path(uid=file_id, version=version_id)
        return FileLink(filename=dest_name, url=url)

    def download(self, *, file_link: Union[str, UUID, FileLink], local_path: Union[str, Path]):
        """
        Download the file associated with a given FileLink to the local computer.

        Parameters
        ----------
        file_link: FileLink, str, UUID
            Resource referencing the file.
        local_path: str, Path
            Path to save file on the local computer. If `local_path` is a directory,
            then the filename of this FileLink object will be appended to the path.

        """
        file_link = self._resolve_file_link(file_link)

        if isinstance(local_path, str):
            directory, filename = os.path.split(local_path)
            if len(filename) == 0:  # the string ended with /
                final_path = Path(directory) / file_link.filename
            else:
                final_path = Path(directory) / filename
        elif local_path.is_dir():
            final_path = local_path / file_link.filename
        else:
            final_path = local_path

        if self._is_external_url(file_link.url):  # Pull it from where ever it lives
            final_url = file_link.url
        elif self._validate_local_url(file_link.url):
            # The "/content-link" route returns a pre-signed url to download the file.
            content_link = self._get_path_from_file_link(file_link, action='content-link')
            content_link_response = self.session.get_resource(content_link)
            pre_signed_url = content_link_response['pre_signed_read_link']
            final_url = rewrite_s3_links_locally(pre_signed_url, self.session.s3_endpoint_url)
        else:  # Unrecognized
            raise ValueError(f"URL was malformed for a local file resource ({file_link.url}).")

        download_response = requests.get(final_url)
        write_file_locally(download_response.content, final_path)

    def read(self, *, file_link: Union[str, UUID, FileLink]):
        """
        Read the file associated with a given FileLink.

        Parameters
        ----------
        file_link: FileLink, str, UUID
            Resource referencing the file.

        Returns
        -------
        I/O stream
            The contents of the file.

        """
        file_link = self._resolve_file_link(file_link)

        if self._is_external_url(file_link.url):  # Pull it from where ever it lives
            final_url = file_link.url
        elif self._validate_local_url(file_link.url):
            # The "/content-link" route returns a pre-signed url to download the file.
            content_link = self._get_path_from_file_link(file_link, action='content-link')
            content_link_response = self.session.get_resource(content_link)
            pre_signed_url = content_link_response['pre_signed_read_link']
            final_url = rewrite_s3_links_locally(pre_signed_url, self.session.s3_endpoint_url)
        else:  # Unrecognized
            raise ValueError(f"URL was malformed for a local file resource ({file_link.url}).")

        download_response = requests.get(final_url)
        return download_response.content

    def process(self, *, file_link: Union[FileLink, str, UUID],
                processing_type: FileProcessingType,
                wait_for_response: bool = True,
                timeout: float = 2 * 60,
                polling_delay: float = 1.0) -> Union[JobSubmissionResponse,
                                                     Dict[FileProcessingType,
                                                          FileProcessingResult]]:
        """
        Start a File Processing async job, returning a pollable job response.

        Parameters
        ----------
        file_link: FileLink, str, or UUID
            The file to process.
        processing_type: FileProcessingType
            The type of file processing to invoke.
        wait_for_response: bool
            Whether to wait for a result, vs. just return a job handle.  Default: True
        timeout: float
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.  Default: 120 seconds.
        polling_delay: float
            How long to delay between each polling retry attempt.  Default: 1 second.

        Returns
        -------
        JobSubmissionResponse
            A JobSubmissionResponse which can be used to poll for the result.

        """
        file_link = self._resolve_file_link(file_link)
        if not self._validate_local_url(file_link.url):
            raise ValueError(f"Only on-platform resources can be processed. "
                             f"Passed URL {file_link.url}.")

        params = {"processing_type": processing_type.value}
        response = self.session.put_resource(
            self._get_path_from_file_link(file_link, action="processed"),
            json={},
            params=params
        )
        job = JobSubmissionResponse.build(response)
        logger.info('Build job submitted with job ID {}.'.format(job.job_id))

        if wait_for_response:
            return self.poll_file_processing_job(file_link=file_link,
                                                 processing_type=processing_type,
                                                 job_id=job.job_id, timeout=timeout,
                                                 polling_delay=polling_delay)
        else:
            return job

    def poll_file_processing_job(self, *, file_link: FileLink,
                                 processing_type: FileProcessingType,
                                 job_id: UUID,
                                 timeout: float = 2 * 60,
                                 polling_delay: float = 1.0) -> Dict[FileProcessingType,
                                                                     FileProcessingResult]:
        """
        [ALPHA] Poll for the result of the file processing task.

        Parameters
        ----------
        file_link: FileLink
            The file to process.
        processing_type: FileProcessingType
            The processing algorithm to apply.
        job_id: UUID
           The background job ID to poll for.
        timeout:
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.
        polling_delay:
            How long to delay between each polling retry attempt.

        Returns
        -------
        None
           This method will raise an appropriate exception if the job failed, else
           it will return None to indicate the job was successful.

        """
        # Poll for job completion - this will raise an error if the job failed
        _poll_for_job_completion(self.session, self.project_id, job_id, timeout=timeout,
                                 polling_delay=polling_delay)

        return self.file_processing_result(file_link=file_link, processing_types=[processing_type])

    def file_processing_result(self, *,
                               file_link: FileLink,
                               processing_types: List[FileProcessingType]) -> \
            Dict[FileProcessingType, FileProcessingResult]:
        """
        Return the file processing result for the given file link and processing type.

        Parameters
        ----------
        file_link: FileLink
            The file to process
        processing_types: FileProcessingType
            A list of the particular file processing types to retrieve

        Returns
        -------
        Map[FileProcessingType, FileProcessingResult]
            The file processing results, mapped by processing type.

        """
        processed_results_path = self._get_path_from_file_link(file_link, action="processed")

        params = []
        for proc_type in processing_types:
            params.append(('processing_type', proc_type.value))

        response = self.session.get_resource(processed_results_path, params=params)
        results_json = response['results']
        results = {}
        for result_json in results_json:
            processing_type = FileProcessingType[result_json['processing_type']]
            data = result_json['data']

            if processing_type == FileProcessingType.VALIDATE_CSV:
                data = CsvValidationData.build(data)

            result = FileProcessingResult(processing_type, data)
            results[processing_type] = result

        return results

    def delete(self, file_link: FileLink):
        """
        Delete the file associated with a given FileLink from the database.

        Parameters
        ----------
        file_link: FileLink
            Resource referencing the external file.

        """
        if self._is_external_url(file_link.url):
            raise ValueError(f"Only local resources can be deleted; passed URL {file_link.url}")
        file_id = self._get_ids_from_url(file_link.url)[0]
        if file_id is None:
            raise ValueError(f"URL was malformed for local resources; passed URL {file_link.url}")
        data = self.session.delete_resource(self._get_path(file_id))
        return Response(body=data)

    def _resolve_file_link(self, identifier: Union[str, UUID, FileLink]) -> FileLink:
        """Generate the FileLink object referenced by the passed argument."""
        if isinstance(identifier, FileLink):  # Passthrough for convenience
            return identifier
        elif isinstance(identifier, str) and self._is_external_url(identifier):
            # Assume it's an absolute URL
            filename = urlparse(identifier).path.split('/')[-1]
            file_dict = {
                'url': identifier,
                'filename': filename,
                'type': GEMDFileLink.typ
            }
            return FileLink.build(file_dict)
        elif isinstance(identifier, str):  # It's either a filename or URL
            file_id, version_id = self._get_ids_from_url(identifier)
            if file_id is None:  # Try it as a name or a stand-alone UID
                return self.get(identifier)
            else:  # We got a file UID (and possibly a version UID) from a URL
                return self.get(uid=file_id, version=version_id)
        elif isinstance(identifier, UUID):  # File UID
            return self.get(uid=identifier)
        else:
            raise TypeError(f"File Link can only be resolved from str, or UUID."
                            f"Instead got {type(identifier)} {identifier}.")

    def _is_external_url(self, url: str):
        """Check if the URL is absolute and not associated with this platform instance."""
        parsed = urlparse(url)
        if len(parsed.scheme) == 0 or len(parsed.netloc) == 0:
            # Relative
            return False

        return urlparse(self._get_path()).netloc != parsed.netloc

    def _validate_local_url(self, url):
        """Verify link is well formed."""
        if self._is_external_url(url):
            return False

        return self._get_ids_from_url(url)[1] is not None  # Implies file_id is None, too



from copy import copy
from typing import List, Union, Optional, Tuple
from uuid import UUID

from deprecation import deprecated
from gemd.entity.object import MaterialRun

from gemd.entity.link_by_uid import LinkByUID
from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument, format_escaped_url
from citrine.resources.data_concepts import CITRINE_SCOPE, _make_link_by_uid
from citrine.resources.process_template import ProcessTemplate
from citrine.gemtables.columns import Column, MeanColumn, IdentityColumn, OriginalUnitsColumn, \
    ConcatColumn
from citrine.gemtables.rows import Row
from citrine.gemtables.variables import Variable, IngredientIdentifierByProcessTemplateAndName, \
    IngredientQuantityByProcessAndName, IngredientQuantityDimension, \
    IngredientIdentifierInOutput, IngredientQuantityInOutput, \
    IngredientLabelsSetByProcessAndName, IngredientLabelsSetInOutput


class TableBuildAlgorithm(BaseEnumeration):
    """[ALPHA] The algorithm to use in automatically building a Table Configuration.

    * SINGLE_ROW corresponds one row per material history
    * FORMULATIONS corresponds to one row per ingredient, intermediate, or terminal
      material, splitting the graph at branches.
    """

    SINGLE_ROW = "single_row"
    FORMULATIONS = "formulations"


class TableConfig(Resource["TableConfig"]):
    """
    [ALPHA] The Table Configuration used to build GEM Tables.

    Parameters
    ----------
    name: str
        Name of the Table Configuration
    description: str
        Description of the Table Configuration
    datasets: list[UUID]
        Datasets that are in scope for the table, as a list of dataset uuids
    variables: list[Variable]
        Variable definitions, which define data from the material histories to use in the columns
    rows: list[Row]
        List of row definitions that define the rows of the table
    columns: list[Column]
        Column definitions, which describe how the variables are shaped into the table

    """

    # FIXME (DML): rename this (this is dependent on the server side)
    _response_key = "ara_definition"
    _resource_type = ResourceTypeEnum.TABLE_DEFINITION

    @staticmethod
    def _get_dups(lst: List) -> List:
        # Hmmn, this looks like a potentially costly operation?!
        return [x for x in lst if lst.count(x) > 1]

    config_uid = properties.Optional(properties.UUID(), 'definition_id')
    """:Optional[UUID]: Unique ID of the table config, independent of its version."""
    version_number = properties.Optional(properties.Integer, 'version_number')
    """:Optional[int]: The version of the table config, starting from 1.
    It increases every time the table config is updated."""
    version_uid = properties.Optional(properties.UUID(), 'id')
    """:Optional[UUID]: Unique ID that specifies one version of one table config."""

    name = properties.String("name")
    description = properties.String("description")
    datasets = properties.List(properties.UUID, "datasets")
    variables = properties.List(properties.Object(Variable), "variables")
    rows = properties.List(properties.Object(Row), "rows")
    columns = properties.List(properties.Object(Column), "columns")

    def __init__(self, name: str, *, description: str, datasets: List[UUID],
                 variables: List[Variable], rows: List[Row], columns: List[Column]):
        self.name = name
        self.description = description
        self.datasets = datasets
        self.rows = rows
        self.variables = variables
        self.columns = columns

        # Note that these validations only apply at construction time. The current intended usage
        # is for this object to be created holistically; if changed, then these will need
        # to move into setters.
        names = [x.name for x in variables]
        dup_names = self._get_dups(names)
        if len(dup_names) > 0:
            raise ValueError("Multiple variables defined these names,"
                             " which much be unique: {}".format(dup_names))
        headers = [x.headers for x in variables]
        dup_headers = self._get_dups(headers)
        if len(dup_headers) > 0:
            raise ValueError("Multiple variables defined these headers,"
                             " which much be unique: {}".format(dup_headers))

        missing_variables = [x.data_source for x in columns if x.data_source not in names]
        if len(missing_variables) > 0:
            raise ValueError("The data_source of the columns must match one of the variable names,"
                             " but {} were missing".format(missing_variables))

    @property
    def uid(self) -> UUID:
        """Unique ID of the table config, independent of its version."""
        return self.config_uid

    @uid.setter
    def uid(self, new_uid: Union[str, UUID]) -> None:
        """Set the unique ID of the table config, independent of its version."""
        self.config_uid = new_uid

    def add_columns(self, *,
                    variable: Variable,
                    columns: List[Column],
                    name: Optional[str] = None,
                    description: Optional[str] = None
                    ) -> 'TableConfig':
        """[ALPHA] Add a variable and one or more columns to this TableConfig (out-of-place).

        This method checks that the variable name is not already in use and that the columns
        only reference that variable.  It is *not* able to check if the columns and the variable
        are compatible (yet, at least).

        Parameters
        ----------
        variable: Variable
            Variable to add and use in the added columns
        columns: list[Column]
            Columns to add, which must only reference the added variable
        name: Optional[str]
            Optional renaming of the table
        description: Optional[str]
            Optional re-description of the table

        """
        if variable.name in [x.name for x in self.variables]:
            raise ValueError("The variable name {} is already used".format(variable.name))

        mismatched_data_source = [x for x in columns if x.data_source != variable.name]
        if len(mismatched_data_source):
            raise ValueError("Column.data_source must be {} but found {}"
                             .format(variable.name, mismatched_data_source))

        new_config = TableConfig(
            name=name or self.name,
            description=description or self.description,
            datasets=copy(self.datasets),
            rows=copy(self.rows),
            variables=copy(self.variables) + [variable],
            columns=copy(self.columns) + columns
        )
        new_config.version_number = copy(self.version_number)
        new_config.config_uid = copy(self.config_uid)
        new_config.version_uid = copy(self.version_uid)
        return new_config

    def add_all_ingredients(self, *,
                            process_template: Union[LinkByUID, ProcessTemplate, str, UUID],
                            project,
                            quantity_dimension: IngredientQuantityDimension,
                            scope: str = CITRINE_SCOPE,
                            unit: Optional[str] = None
                            ):
        """[ALPHA] Add variables and columns for all of the possible ingredients in a process.

        For each allowed ingredient name in the process template there is a column for the id of
        the ingredient, id for ingredient labels, and a column for the quantity of the ingredient.
        If the quantities are given in absolute amounts then there is also a column for units.

        Parameters
        ------------
        process_template: Union[LinkByUID, ProcessTemplate, str, UUID]
            representation of a registered process template
        project: Project
            a project that has access to the process template
        quantity_dimension: IngredientQuantityDimension
            the dimension in which to report ingredient quantities
        scope: Optional[str]
            the scope for which to get ingredient ids (default is Citrine scope, 'id')
        unit: Optional[str]
            the units for the quantity, if selecting Absolute Quantity

        """
        dimension_display = {
            IngredientQuantityDimension.ABSOLUTE: "absolute quantity",
            IngredientQuantityDimension.MASS: "mass fraction",
            IngredientQuantityDimension.VOLUME: "volume fraction",
            IngredientQuantityDimension.NUMBER: "number fraction"
        }
        link = _make_link_by_uid(process_template)
        process: ProcessTemplate = project.process_templates.get(uid=link)
        if not process.allowed_names:
            raise RuntimeError(
                "Cannot add ingredients for process template \'{}\' because it has no defined "
                "ingredients (allowed_names is not defined).".format(process.name))

        new_variables = []
        new_columns = []
        for name in process.allowed_names:
            identifier_variable = IngredientIdentifierByProcessTemplateAndName(
                name='_'.join([process.name, name, str(hash(link.id + name + scope))]),
                headers=[process.name, name, scope],
                process_template=link,
                ingredient_name=name,
                scope=scope
            )
            quantity_variable = IngredientQuantityByProcessAndName(
                name='_'.join([process.name, name, str(hash(
                    link.id + name + dimension_display[quantity_dimension]))]),
                headers=[process.name, name, dimension_display[quantity_dimension]],
                process_template=link,
                ingredient_name=name,
                quantity_dimension=quantity_dimension,
                unit=unit
            )
            label_variable = IngredientLabelsSetByProcessAndName(
                name='_'.join([process.name, name, str(hash(
                    link.id + name + 'Labels'))]),
                headers=[process.name, name, 'Labels'],
                process_template=link,
                ingredient_name=name,
            )

            if identifier_variable.name not in [var.name for var in self.variables]:
                new_variables.append(identifier_variable)
                new_columns.append(IdentityColumn(data_source=identifier_variable.name))
            new_variables.append(quantity_variable)
            new_columns.append(MeanColumn(data_source=quantity_variable.name))
            if quantity_dimension == IngredientQuantityDimension.ABSOLUTE:
                new_columns.append(OriginalUnitsColumn(data_source=quantity_variable.name))
            if label_variable.name not in [var.name for var in self.variables]:
                new_variables.append(label_variable)
                new_columns.append(
                    ConcatColumn(
                        data_source=label_variable.name,
                        subcolumn=IdentityColumn(data_source=label_variable.name)
                    )
                )

        new_config = TableConfig(
            name=self.name,
            description=self.description,
            datasets=copy(self.datasets),
            rows=copy(self.rows),
            variables=copy(self.variables) + new_variables,
            columns=copy(self.columns) + new_columns
        )
        new_config.version_number = copy(self.version_number)
        new_config.config_uid = copy(self.config_uid)
        new_config.version_uid = copy(self.version_uid)
        return new_config

    def add_all_ingredients_in_output(self, *,
                                      process_templates: List[LinkByUID],
                                      project,
                                      quantity_dimension: IngredientQuantityDimension,
                                      scope: str = CITRINE_SCOPE,
                                      unit: Optional[str] = None
                                      ):
        """[ALPHA] Add variables and columns for all possible ingredients in a list of processes.

        For each allowed ingredient name in the union of all passed process templates there is a
        column for the id of the ingredient and a column for the quantity of the ingredient.
        Columns are filled with the "InOutput" method halting at any of the passed process
        templates. If the quantities are given in absolute amounts then there is also a column for
        units.

        Parameters
        ------------
        process_templates: List[LinkByUID]
            registered process templates from which to pull allowed ingredients and at which to
            halt searching
        project: Project
            a project that has access to the process template
        quantity_dimension: IngredientQuantityDimension
            the dimension in which to report ingredient quantities
        scope: Optional[str]
            the scope for which to get ingredient ids (default is Citrine scope, 'id')
        unit: Optional[str]
            the units for the quantity, if selecting Absolute Quantity

        """
        dimension_display = {
            IngredientQuantityDimension.ABSOLUTE: "absolute quantity",
            IngredientQuantityDimension.MASS: "mass fraction",
            IngredientQuantityDimension.VOLUME: "volume fraction",
            IngredientQuantityDimension.NUMBER: "number fraction"
        }
        union_allowed_names = []
        for process_template_link in process_templates:
            process: ProcessTemplate = project.process_templates.get(process_template_link)
            if not process.allowed_names:
                raise RuntimeError(
                    f"Cannot add ingredients for process template '{process.name}' "
                    "because it has no defined ingredients (allowed_names is not defined)"
                )
            else:
                union_allowed_names = list(set(union_allowed_names) | set(process.allowed_names))

        new_variables = []
        new_columns = []
        for name in union_allowed_names:
            identifier_variable = IngredientIdentifierInOutput(
                name='_'.join([name, str(hash(name + scope))]),
                headers=[name, scope],
                process_templates=process_templates,
                ingredient_name=name,
                scope=scope
            )
            quantity_variable = IngredientQuantityInOutput(
                name='_'.join([name, str(hash(name + dimension_display[quantity_dimension]))]),
                headers=[name, dimension_display[quantity_dimension]],
                process_templates=process_templates,
                ingredient_name=name,
                quantity_dimension=quantity_dimension,
                unit=unit
            )
            label_variable = IngredientLabelsSetInOutput(
                name='_'.join([name, str(hash(name + 'Labels'))]),
                headers=[name, 'Labels'],
                process_templates=process_templates,
                ingredient_name=name,
            )

            if identifier_variable.name not in [var.name for var in self.variables]:
                new_variables.append(identifier_variable)
                new_columns.append(IdentityColumn(data_source=identifier_variable.name))
            new_variables.append(quantity_variable)
            new_columns.append(MeanColumn(data_source=quantity_variable.name))
            if quantity_dimension == IngredientQuantityDimension.ABSOLUTE:
                new_columns.append(OriginalUnitsColumn(data_source=quantity_variable.name))
            if label_variable.name not in [var.name for var in self.variables]:
                new_variables.append(label_variable)
                new_columns.append(
                    ConcatColumn(
                        data_source=label_variable.name,
                        subcolumn=IdentityColumn(data_source=label_variable.name)
                    )
                )

        new_config = TableConfig(
            name=self.name,
            description=self.description,
            datasets=copy(self.datasets),
            rows=copy(self.rows),
            variables=copy(self.variables) + new_variables,
            columns=copy(self.columns) + new_columns
        )
        new_config.version_number = copy(self.version_number)
        new_config.config_uid = copy(self.config_uid)
        new_config.version_uid = copy(self.version_uid)
        return new_config


class TableConfigCollection(Collection[TableConfig]):
    """[ALPHA] Represents the collection of all Table Configs associated with a project."""

    # FIXME (DML): use newly named properties when they're available
    _path_template = 'projects/{project_id}/ara-definitions'
    _collection_key = 'definitions'
    _resource = TableConfig

    # NOTE: This isn't actually an 'individual key' - both parts (version and
    # definition) are necessary
    _individual_key = None

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def get(self, uid: Union[UUID, str], *, version: Optional[int] = None):
        """[ALPHA] Get a table config.

        If no version is specified, then the most recent version is returned.

        """
        if uid is None:
            raise ValueError("Cannot get when uid=None.  Are you using a registered resource?")
        if version is not None:
            path = self._get_path(uid) + format_escaped_url("/versions/{}", version)
            data = self.session.get_resource(path)
        else:
            path = self._get_path(uid)
            data = self.session.get_resource(path)
            version_numbers = [version_data['version_number'] for version_data in data['versions']]
            index = version_numbers.index(max(version_numbers))
            data['version'] = data['versions'][index]
        return self.build(data)

    def get_for_table(self, table: "GemTable") -> TableConfig:  # noqa: F821
        """
        Get the TableConfig used to build the given table.

        Parameters
        ----------
        table: GemTable
            Table for which to get the config.

        Returns
        -------
        TableConfig
            The table config used to produce the given table.

        """
        # the route to fetch the config is built off the display table route tree
        path = 'projects/{project_id}/display-tables/{uid}/versions/{version}/definition'.format(
            project_id=self.project_id, uid=table.uid, version=table.version)
        data = self.session.get_resource(path)
        return self.build(data)

    @deprecated(deprecated_in="0.124.0", removed_in="2.0.0",
                details="get_with_version() is deprecated in favor of get()")
    def get_with_version(self, *, table_config_uid: Union[UUID, str],
                         version_number: int) -> TableConfig:
        """[ALPHA] Get a Table Config at a specific version."""
        return self.get(uid=table_config_uid, version=version_number)  # pragma: no cover

    def build(self, data: dict) -> TableConfig:
        """[ALPHA] Build an individual Table Config from a dictionary."""
        version_data = data['version']
        table_config = TableConfig.build(version_data['ara_definition'])
        table_config.version_number = version_data['version_number']
        table_config.version_uid = version_data['id']
        table_config.config_uid = data['definition']['id']
        table_config.project_id = self.project_id
        table_config.session = self.session
        return table_config

    def default_for_material(
            self, *,
            material: Union[MaterialRun, LinkByUID, str, UUID],
            name: str,
            description: str = None,
            algorithm: Optional[TableBuildAlgorithm] = None,
            scope: str = None
    ) -> Tuple[TableConfig, List[Tuple[Variable, Column]]]:
        """
        [ALPHA] Build best-guess default table config for provided terminal material's history.

        Currently generates variables for each templated attribute in the material history in
        either AttributeByTemplate, or if possible AttributeByTemplateAndObjectTemplate.
        Attributes on object templates used in the history are included regardless of their
        presence on data objects in the history. Additionally, each quantity dimension specified on
        ingredients in the material history will be captured in IngredientQuantityByProcessAndName.

        If a generated variable would match ambiguously on the given material history, it is
        excluded from the generated config and included in the second part of the returned tuple.

        Parameters
        ----------
        material: Union[MaterialRun, LinkByUid, str, UUID]
            The terminal material whose history is used to construct a table config.
        name: str
            The name for the table config.
        description: str, optional
            The description of the table config. Defaults to autogenerated message.
        algorithm: TableBuildAlgorithm, optional
            The algorithm to use in generating a Table Configuration from the sample material
            history.  If unspecified, uses the webservice's default.
        scope: str, optional
            [DEPRECATED] Use a LinkByUID instead.
            The scope of the material id. If a uid is provided and scope is not specified,
            then the uid is assumed to be a Citrine ID.

        Returns
        -------
        List[Tuple[Variable, Column]]
            A table config as well as addition variables/columns which would result in
            ambiguous matches if included in the config.


        """
        link = _make_link_by_uid(material, scope)
        params = {
            'id': link.id,
            'scope': link.scope,
            'name': name,
        }
        if description is not None:
            params['description'] = description
        if algorithm is not None:
            if isinstance(algorithm, TableBuildAlgorithm):
                params['algorithm'] = algorithm.value
            else:  # Not per spec, but be forgiving
                params['algorithm'] = str(algorithm)
        data = self.session.get_resource(
            format_escaped_url('projects/{}/table-configs/default', self.project_id),
            params=params,
        )
        config = TableConfig.build(data['config'])
        ambiguous = [(Variable.build(v), Column.build(c)) for v, c in data['ambiguous']]
        return config, ambiguous

    def preview(self, *,
                table_config: TableConfig,
                preview_materials: List[LinkByUID] = None,
                preview_roots: List[LinkByUID] = None
                ) -> dict:
        """[ALPHA] Preview a Table Config on an explicit set of terminal materials.

        Parameters
        ----------
        table_config: TableConfig
            Table Config to preview
        preview_materials: List[LinkByUID]
            List of links to the material runs to use as terminal materials in the preview
        preview_roots: List[LinkByUID]
            [DEPRECATED] Use preview_materials instead

        """
        preview_materials = migrate_deprecated_argument(preview_materials, "preview_materials",
                                                        preview_roots, "preview_roots")
        path = self._get_path() + "/preview"
        body = {
            "definition": table_config.dump(),
            "rows": [x.as_dict() for x in preview_materials]
        }
        return self.session.post_resource(path, body)

    def register(self, table_config: TableConfig) -> TableConfig:
        """[ALPHA] Register a Table Config.

        If the provided TableConfig does not have a definition_uid, create a new element of the
        TableConfigCollection by registering the provided TableConfig. If the provided
        TableConfig does have a uid, update (replace) the TableConfig at that uid with the
        provided TableConfig.

        :param table_config: The TableConfig to register

        :return: The registered TableConfig with updated metadata

        TODO: Consider validating that a resource exists at the given uid before updating.
            The code to do so is not yet implemented on the backend
        """
        # TODO: This is dumping our TableConfig (which encapsulates both
        #  the table config properties, versioned table config properties, as well as the
        #  underlying table config JSON blob) into the Table Config's JSON blob ('definition')
        #  - probably not ideal.
        body = {"definition": table_config.dump()}
        if table_config.config_uid is None:
            data = self.session.post_resource(self._get_path(), body)
            data = data[self._individual_key] if self._individual_key else data
            return self.build(data)
        else:
            # Implement update as a part of register both because:
            # 1) The validation requirements are the same for updating and registering an
            #    TableConfig
            # 2) This prevents users from accidentally registering duplicate Table Configs
            data = self.session.put_resource(self._get_path(table_config.config_uid), body)
            data = data[self._individual_key] if self._individual_key else data
            return self.build(data)

    def update(self, table_config: TableConfig) -> TableConfig:
        """
        [ALPHA] Update a Table Config.

        If the provided Table Config does have a uid, update (replace) the Table Config at that
        uid with the provided TableConfig.

        Raise a ValueError if the provided Table Config does not have a config_uid.

        :param table_config: The Table Config to updated
        :return: The updated Table Config with updated metadata
        """
        if table_config.config_uid is None:
            raise ValueError("Cannot update Table Config without a config_uid."
                             " Please either use register() to initially register this"
                             " Table Config or retrieve the registered details before calling"
                             " update()")
        return self.register(table_config)

    def delete(self, uid: Union[UUID, str]):
        """Table configs cannot be deleted at this time."""
        raise NotImplementedError("Table configs cannot be deleted at this time.")



from itertools import chain
from typing import Iterator, Mapping, Union, List
from uuid import UUID

from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.bounds import RealBounds, CategoricalBounds, MolecularStructureBounds, \
    IntegerBounds, CompositionBounds
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.template.has_property_templates import HasPropertyTemplates
from gemd.entity.template.has_condition_templates import HasConditionTemplates
from gemd.entity.template.has_parameter_templates import HasParameterTemplates
from gemd.entity.value import EmpiricalFormula
from gemd.util import recursive_flatmap, set_uuids

from citrine.builders.auto_configure import AutoConfigureMode
from citrine.informatics.descriptors import RealDescriptor, CategoricalDescriptor, \
    MolecularStructureDescriptor, Descriptor, ChemicalFormulaDescriptor, IntegerDescriptor
from citrine.resources.data_concepts import DataConceptsCollection
from citrine.resources.material_run import MaterialRun
from citrine.resources.project import Project


class NoEquivalentDescriptorError(ValueError):
    """Error that is raised when the bounds in a template have no equivalent descriptor."""

    pass


def template_to_descriptor(template: AttributeTemplate, *,
                           headers: List[str] = []) -> Descriptor:
    """
    Convert a GEMD attribute template into an AI Engine Descriptor.

    IntBounds cannot be converted because they have no matching descriptor type.
    CompositionBounds can only be converted when every component is an element, in which case
    they are converted to ChemicalFormulaDescriptors.

    Parameters
    ----------
    template: AttributeTemplate
        Template to convert into a descriptor
    headers: List[str]
        Names of parent relationships to includes as prefixes
        to the template name in the descriptor key
        Default: []

    Returns
    -------
    Descriptor
        Descriptor with a key matching the template name and type corresponding to the bounds

    """
    headers = headers + [template.name]
    descriptor_key = '~'.join(headers)
    bounds = template.bounds
    if isinstance(bounds, RealBounds):
        return RealDescriptor(
            key=descriptor_key,
            lower_bound=bounds.lower_bound,
            upper_bound=bounds.upper_bound,
            units=bounds.default_units
        )
    if isinstance(bounds, CategoricalBounds):
        return CategoricalDescriptor(
            key=descriptor_key,
            categories=bounds.categories
        )
    if isinstance(bounds, MolecularStructureBounds):
        return MolecularStructureDescriptor(
            key=descriptor_key
        )
    if isinstance(bounds, CompositionBounds):
        if set(bounds.components).issubset(EmpiricalFormula.all_elements()):
            return ChemicalFormulaDescriptor(
                key=descriptor_key
            )
        else:
            msg = "Cannot create descriptor for CompositionBounds with non-atomic components"
            raise NoEquivalentDescriptorError(msg)
    if isinstance(bounds, IntegerBounds):
        return IntegerDescriptor(
            key=descriptor_key,
            lower_bound=bounds.lower_bound,
            upper_bound=bounds.upper_bound
        )
    raise ValueError("Template has unrecognized bounds: {}".format(type(bounds)))


class PlatformVocabulary(Mapping[str, Descriptor]):
    """
    Dictionary of descriptors that define a controlled vocabulary for the AI Engine.

    Parameters
    ----------
    entries: Mapping[str, Descriptor]
        Entries in the dictionary, indexed by a convenient name.
        To build from templates, use PlatformVocabulary.from_templates
        To build from a material, use PlatformVocabulary.from_material

    """

    def __init__(self, *, entries: Mapping[str, Descriptor]):
        self._entries = entries

    def __getitem__(self, k: str) -> Descriptor:
        return self._entries[k]

    def __len__(self):
        return len(self._entries)

    def __iter__(self) -> Iterator[str]:
        return iter(self._entries)

    @staticmethod
    def from_templates(*, project: Project, scope: str):
        """
        Build a PlatformVocabulary from the templates visible to a project.

        All of the templates with the given scope are downloaded and converted into descriptors.
        The uid values associated with that scope are used as the index into the dictionary.
        For example, using scope "my_templates" with a template with
        uids={"my_templates": "density"} would be indexed into the dictionary as "density".

        Parameters
        ----------
        project: Project
            Project on the Citrine Platform to read templates from
        scope: str
            Unique ID scope from which to pull the template names

        Returns
        -------
        PlatformVocabulary

        """
        def _from_collection(collection: DataConceptsCollection):
            return {x.uids[scope]: x for x in collection.list() if scope in x.uids}

        properties = _from_collection(project.property_templates)
        parameters = _from_collection(project.parameter_templates)
        conditions = _from_collection(project.condition_templates)

        res = {}
        for k, v in chain(properties.items(), parameters.items(), conditions.items()):
            try:
                desc = template_to_descriptor(v)
                res[k] = desc
            except NoEquivalentDescriptorError:
                continue

        return PlatformVocabulary(entries=res)

    @staticmethod
    def from_material(
            *,
            project: Project,
            material: Union[str, UUID, LinkByUID, MaterialRun],
            mode: AutoConfigureMode = AutoConfigureMode.PLAIN,
            full_history: bool = True
    ):
        """[ALPHA] Build a PlatformVocabulary from templates appearing in a material history.

        All of the attribute templates that appear throughout the material's history
        are extracted and converted into descriptors.

        Descriptor keys are formatted according to the option set by mode.
        For example, if a condition template with name 'Condition 1'
        appears in a parent process with name 'Parent',
        the mode option produces the following descriptor key:

        mode = AutoConfigMode.PLAIN       --> 'Parent~Condition 1'
        mode = AutoConfigMode.FORMULATION --> 'Condition 1'

        Parameters
        ----------
        project: Project
            Project to use when accessing the Citrine Platform.
        material: Union[str, UUID, LinkByUID, MaterialRun]
            A representation of the material to extract descriptors from.
        mode: AutoConfigureMode
            Formatting option for descriptor keys in the platform vocabulary.
            Option AutoConfigMode.PLAIN includes headers from the parent object,
            whereas option AutoConfigMode.FORMULATION does not.
            Default: AutoConfigureMode.PLAIN
        full_history: bool
            Whether to extract descriptors from the full material history,
            or only the provided (terminal) material.
            Default: True

        Returns
        -------
        PlatformVocabulary

        """
        if not isinstance(mode, AutoConfigureMode):
            raise TypeError('mode must be an option from AutoConfigureMode')

        # Full history not needed when full_history = False
        # But is convenient to populate templates for terminal material
        history = project.material_runs.get_history(id=material)
        if full_history:
            search_history = recursive_flatmap(history, lambda x: [x], unidirectional=False)
            set_uuids(search_history, 'id')
        else:
            # Limit the search to contain the terminal material/process/measurements
            search_history = [history.spec.template, history.process.template]
            search_history.extend([msr.template for msr in history.measurements])
            search_history = [x for x in search_history if x is not None]  # Edge case safety

        # Extract templates and formatted keys
        res = {}
        for obj in search_history:
            # Extract all templates
            templates = []
            if isinstance(obj, HasPropertyTemplates):
                for property in obj.properties:
                    templates.append(property[0])
            if isinstance(obj, HasConditionTemplates):
                for condition in obj.conditions:
                    templates.append(condition[0])
            if isinstance(obj, HasParameterTemplates):
                for parameter in obj.parameters:
                    templates.append(parameter[0])

            # Assemble to descriptors
            headers = []
            if mode == AutoConfigureMode.PLAIN:
                headers.append(obj.name)

            for tmpl in templates:
                try:
                    desc = template_to_descriptor(tmpl, headers=headers)
                    res[desc.key] = desc
                except NoEquivalentDescriptorError:
                    continue

        return PlatformVocabulary(entries=res)



from deprecation import deprecated
from typing import List

from citrine.resources.status_detail import StatusDetail
from citrine._serialization import properties


class AIResourceMetadata():
    """Abstract class for representing common metadata for Resources."""

    created_by = properties.Optional(properties.UUID, 'created_by', serializable=False)
    """:Optional[UUID]: id of the user who created the resource"""
    create_time = properties.Optional(properties.Datetime, 'create_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was created"""

    updated_by = properties.Optional(properties.UUID, 'updated_by', serializable=False)
    """:Optional[UUID]: id of the user who most recently updated the resource,
    if it has been updated"""
    update_time = properties.Optional(properties.Datetime, 'update_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was most recently updated,
    if it has been updated"""

    archived = properties.Boolean('archived', default=False)
    """:bool: whether the resource is archived (hidden but not deleted)"""
    archived_by = properties.Optional(properties.UUID, 'archived_by', serializable=False)
    """:Optional[UUID]: id of the user who archived the resource, if it has been archived"""
    archive_time = properties.Optional(properties.Datetime, 'archive_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was archived,
    if it has been archived"""

    status = properties.Optional(properties.String(), 'status', serializable=False)
    """:Optional[str]: short description of the resource's status"""
    status_info = properties.Optional(
        properties.List(properties.String()),
        'status_info',
        serializable=False
    )

    """:Optional[List[str]]: human-readable explanations of the status"""
    status_detail = properties.List(properties.Object(StatusDetail), 'status_detail', default=[],
                                    serializable=False)
    """:List[StatusDetail]: a list of structured status info, containing the message and level"""

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental(self) -> bool:
        """[DEPRECATED] whether the execution is experimental (newer, less well-tested)."""  # noqa - insisting this docstring is a signature
        return False

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental_reasons(self) -> List[str]:
        """[DEPRECATED] human-readable reasons why the execution is experimental."""
        return []



from typing import Type, Set

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.predictor_evaluation_metrics import PredictorEvaluationMetric
from citrine.informatics.predictor_evaluator import PredictorEvaluator, HoldoutSetEvaluator,\
    CrossValidationEvaluator

__all__ = ['MetricValue',
           'RealMetricValue',
           'PredictedVsActualRealPoint',
           'PredictedVsActualCategoricalPoint',
           'RealPredictedVsActual',
           'CategoricalPredictedVsActual',
           'ResponseMetrics',
           'PredictorEvaluationResult',
           'CrossValidationResult']


class MetricValue(PolymorphicSerializable["MetricValue"]):
    """Value associated with a metric computed during a Predictor Evaluation Workflow."""

    def __init__(self):
        """These are results, so they should be built rather than constructed."""
        pass  # pragma: no cover

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            "RealMetricValue": RealMetricValue,
            "RealPredictedVsActual": RealPredictedVsActual,
            "CategoricalPredictedVsActual": CategoricalPredictedVsActual
        }[data["type"]]


class RealMetricValue(Serializable["RealMetricValue"], MetricValue):
    """Mean and standard error computed for a real-valued metric."""

    mean = properties.Float("mean")
    """:float: Mean value"""
    standard_error = properties.Optional(properties.Float(), "standard_error")
    """:Optional[float]: Standard error of the mean"""
    typ = properties.String('type', default='RealMetricValue', deserializable=False)

    def __eq__(self, other):
        if isinstance(other, RealMetricValue):
            return self.mean == other.mean and self.standard_error == other.standard_error
        else:
            return False


class PredictedVsActualRealPoint(Serializable["PredictedVsActualRealPoint"]):
    """Predicted vs. actual data for a single real-valued data point."""

    uuid = properties.UUID("uuid")
    """:UUID: Unique Citrine id given to the candidate"""
    identifiers = properties.Set(properties.String, "identifiers")
    """:Set[str]: Set of globally unique identifiers given to the candidate"""
    trial = properties.Integer("trial")
    """:int: 1-based index of the trial this candidate belonged to"""
    fold = properties.Integer("fold")
    """:int: 1-based index of the fold this candidate belonged to"""
    predicted = properties.Object(RealMetricValue, "predicted")
    """:RealMetricValue: Predicted value"""
    actual = properties.Object(RealMetricValue, "actual")
    """:RealMetricValue: Actual value"""

    def __init__(self):
        pass  # pragma: no cover


class PredictedVsActualCategoricalPoint(Serializable["PredictedVsActualCategoricalPoint"]):
    """Predicted vs. actual data for a single categorical data point."""

    uuid = properties.UUID("uuid")
    """:UUID: Unique Citrine id given to the candidate"""
    identifiers = properties.Set(properties.String, "identifiers")
    """:Set[str]: Set of globally unique identifiers given to the candidate"""
    trial = properties.Integer("trial")
    """:int: 1-based index of the trial this candidate belonged to"""
    fold = properties.Integer("fold")
    """:int: 1-based index of the fold this candidate belonged to"""
    predicted = properties.Mapping(properties.String, properties.Float, "predicted")
    """:Dict[str, float]: Predicted class probabilities defined as a map from each class name
    to its relative frequency"""
    actual = properties.Mapping(properties.String, properties.Float, "actual")
    """:Dict[str, float]: Actual class probabilities defined as a map from each class name
    to its relative frequency"""

    def __init__(self):
        pass  # pragma: no cover


class CategoricalPredictedVsActual(Serializable["CategoricalPredictedVsActual"], MetricValue):
    """List of predicted vs. actual data points for a categorical value."""

    value = properties.List(properties.Object(PredictedVsActualCategoricalPoint), "value")
    """:List[PredictedVsActualCategoricalPoint]: List of predicted vs. actual data computed during
    a predictor evaluation. This is a flattened list that contains data for all
    trials and folds."""
    typ = properties.String('type', default='CategoricalPredictedVsActual', deserializable=False)

    def __iter__(self):
        return iter(self.value)

    def __getitem__(self, item: int):
        return self.value[item]


class RealPredictedVsActual(Serializable["RealPredictedVsActual"], MetricValue):
    """List of predicted vs. actual data points for a real value."""

    value = properties.List(properties.Object(PredictedVsActualRealPoint), "value")
    """:List[PredictedVsActualRealPoint]: List of predicted vs. actual data computed during
    a predictor evaluation. This is a flattened list that contains data for all
    trials and folds."""
    typ = properties.String('type', default='RealPredictedVsActual', deserializable=False)

    def __iter__(self):
        return iter(self.value)

    def __getitem__(self, item: int):
        return self.value[item]


class ResponseMetrics(Serializable["ResponseMetrics"]):
    """Set of metrics computed by a Predictor Evaluator for a single response.

    Results computed for a metric can be accessed by the metric's ``__repr__`` or
    by the metric itself.

    """

    metrics = properties.Mapping(properties.String, properties.Object(MetricValue), "metrics")
    """:Dict[str, MetricValue]: Metrics computed for a single response, keyed by the
    metric's ``__repr__``."""

    def __init__(self):
        pass  # pragma: no cover

    def __iter__(self):
        return iter(self.metrics)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.metrics[item]
        elif isinstance(item, PredictorEvaluationMetric):
            return self.metrics[repr(item)]
        else:
            raise TypeError("Cannot index ResponseMetrics with a {}".format(type(item)))


class PredictorEvaluationResult(PolymorphicSerializable["PredictorEvaluationResult"]):
    """A Citrine Predictor Evaluation Result.

    This class represents a set of metrics computed by a Predictor Evaluator.
    """

    def __init__(self):
        pass  # pragma: no cover

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            "CrossValidationResult": CrossValidationResult,
            "HoldoutSetResult": HoldoutSetResult
        }[data["type"]]

    @property
    def evaluator(self) -> PredictorEvaluator:
        """:PredictorEvaluator: Evaluator that produced the result."""
        raise NotImplementedError  # pragma: no cover

    @property
    def responses(self) -> Set[str]:
        """Predictor responses that were evaluated."""
        raise NotImplementedError  # pragma: no cover

    @property
    def metrics(self) -> Set[PredictorEvaluationMetric]:
        """:Set[PredictorEvaluationMetric]: Metrics computed for predictor responses."""
        raise NotImplementedError  # pragma: no cover


class CrossValidationResult(Serializable["CrossValidationResult"], PredictorEvaluationResult):
    """Result of performing a cross-validation evaluation on a predictor.

    Results for a cross-validated response can be accessed via ``cvResult['response_name']``,
    where ``cvResult`` is a
    :class:`citrine.informatics.predictor_evaluation_result.CrossValidationResult`
    and ``'response_name'`` is a response analyzed by a
    :class:`citrine.informatics.predictor_evaluator.CrossValidationEvaluator`.

    """

    _evaluator = properties.Object(CrossValidationEvaluator, "evaluator")
    _response_results = properties.Mapping(properties.String, properties.Object(ResponseMetrics),
                                           "response_results")
    typ = properties.String('type', default='CrossValidationResult', deserializable=False)

    def __getitem__(self, item):
        return self._response_results[item]

    def __iter__(self):
        return iter(self.responses)

    @property
    def evaluator(self) -> CrossValidationEvaluator:
        """:PredictorEvaluator: Evaluator that produced this result."""
        return self._evaluator

    @property
    def responses(self) -> Set[str]:
        """Responses for which results are present."""
        return set(self._response_results.keys())

    @property
    def metrics(self) -> Set[PredictorEvaluationMetric]:
        """:Set[PredictorEvaluationMetric]: Metrics for which results are present."""
        return self._evaluator.metrics


class HoldoutSetResult(Serializable["HoldoutSetResult"], PredictorEvaluationResult):
    """Result of performing holdout evaluation on a predictor.

    Results held-out response can be accessed via ``result['response_name']``,
    where ``result`` is a
    :class:`citrine.informatics.predictor_evaluation_result.HoldoutSetResult`
    and ``'response_name'`` is a response analyzed by a
    :class:`citrine.informatics.predictor_evaluator.HoldoutSetEvaluator`.

    """

    _evaluator = properties.Object(HoldoutSetEvaluator, "evaluator")
    _response_results = properties.Mapping(properties.String, properties.Object(ResponseMetrics),
                                           "response_results")
    typ = properties.String('type', default='HoldoutSetResult', deserializable=False)

    def __getitem__(self, item):
        return self._response_results[item]

    def __iter__(self):
        return iter(self.responses)

    @property
    def evaluator(self) -> HoldoutSetEvaluator:
        """:PredictorEvaluator: Evaluator that produced this result."""
        return self._evaluator

    @property
    def responses(self) -> Set[str]:
        """Responses for which results are present."""
        return set(self._response_results.keys())

    @property
    def metrics(self) -> Set[PredictorEvaluationMetric]:
        """:Set[PredictorEvaluationMetric]: Metrics for which results are present."""
        return self._evaluator.metrics



from typing import Set

from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.descriptors import FormulationDescriptor
from citrine.informatics.predictors import Predictor

__all__ = ['LabelFractionsPredictor']


class LabelFractionsPredictor(
        VersionedEngineResource['LabelFractionsPredictor'], Predictor):
    """A predictor interface that computes the relative proportions of labeled ingredients.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        description of the predictor
    input_descriptor: FormulationDescriptor
        descriptor that contains formulation data
    labels: Set[str]
        labels to compute the quantity fractions of

    """

    input_descriptor = _properties.Object(FormulationDescriptor, 'data.instance.input')
    labels = _properties.Set(_properties.String, 'data.instance.labels')

    typ = _properties.String('data.instance.type', default='LabelFractions',
                             deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 input_descriptor: FormulationDescriptor,
                 labels: Set[str]):
        self.name: str = name
        self.description: str = description
        self.input_descriptor: FormulationDescriptor = input_descriptor
        self.labels: Set[str] = labels

    def __str__(self):
        return '<LabelFractionsPredictor {!r}>'.format(self.name)



"""Resources that represent material run data objects."""
import json
import os
from logging import getLogger
from typing import List, Dict, Optional, Type, Iterator, Union
from uuid import UUID

from citrine._rest.resource import Resource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String, LinkOrElse, Mapping, Object
from citrine._utils.functions import format_escaped_url
from citrine.resources.data_concepts import DataConcepts, _make_link_by_uid
from citrine.resources.material_spec import MaterialSpecCollection
from citrine.resources.object_runs import ObjectRun, ObjectRunCollection
from gemd.entity.file_link import FileLink
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.object.material_run import MaterialRun as GEMDMaterialRun
from gemd.entity.object.material_spec import MaterialSpec as GEMDMaterialSpec
from gemd.entity.template.material_template import MaterialTemplate as GEMDMaterialTemplate
from gemd.entity.object.process_run import ProcessRun as GEMDProcessRun
from gemd.json import GEMDEncoder
from gemd.util import writable_sort_order

logger = getLogger(__name__)


class MaterialRun(ObjectRun, Resource['MaterialRun'], GEMDMaterialRun):
    """
    A material run.

    Parameters
    ----------
    name: str
        Name of the material run.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the material run.
    process: ProcessRun
        Process that produces this material.
    sample_type: str, optional
        The form of this sample. Optionals are "experimental", "virtual", "production", or
        "unknown." Default is "unknown."
    spec: MaterialSpec
        The material specification of which this is an instance.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    Attributes
    ----------
    measurements: List[MeasurementRun], optional
        Measurements performed on this material. The link is established by creating the
        measurement run and settings its `material` field to this material run.

    """

    _response_key = GEMDMaterialRun.typ  # 'material_run'

    name = String('name', override=True)
    uids = Mapping(String('scope'), String('id'), 'uids', override=True)
    tags = PropertyOptional(PropertyList(String()), 'tags', override=True)
    notes = PropertyOptional(String(), 'notes', override=True)
    process = PropertyOptional(LinkOrElse(), 'process', override=True)
    sample_type = String('sample_type', override=True)
    spec = PropertyOptional(LinkOrElse(), 'spec', override=True)
    file_links = PropertyOptional(PropertyList(Object(FileLink)), 'file_links', override=True)
    typ = String('type')

    def __init__(self,
                 name: str,
                 *,
                 uids: Optional[Dict[str, str]] = None,
                 tags: Optional[List[str]] = None,
                 notes: Optional[str] = None,
                 process: Optional[GEMDProcessRun] = None,
                 sample_type: Optional[str] = "unknown",
                 spec: Optional[GEMDMaterialSpec] = None,
                 file_links: Optional[List[FileLink]] = None):
        if uids is None:
            uids = dict()
        DataConcepts.__init__(self, GEMDMaterialRun.typ)
        GEMDMaterialRun.__init__(self, name=name, uids=uids,
                                 tags=tags, process=process,
                                 sample_type=sample_type, spec=spec,
                                 file_links=file_links, notes=notes)

    def __str__(self):
        return '<Material run {!r}>'.format(self.name)


class MaterialRunCollection(ObjectRunCollection[MaterialRun]):
    """Represents the collection of all material runs associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/material-runs'
    _dataset_agnostic_path_template = 'projects/{project_id}/material-runs'
    _individual_key = 'material_run'
    _collection_key = 'material_runs'
    _resource = MaterialRun

    @classmethod
    def get_type(cls) -> Type[MaterialRun]:
        """Return the resource type in the collection."""
        return MaterialRun

    def get_history(self, *, id: Union[str, UUID, LinkByUID, MaterialRun],
                    scope: Optional[str] = None) -> Type[MaterialRun]:
        """
        Get the history associated with a terminal material.

        The history contains every single every process, ingredient and material that went into
        the terminal material as well as the measurements that were performed on all of those
        materials. The returned object is a material run with all of its fields fully populated.

        Parameters
        ----------
        id: Union[UUID, str, LinkByUID, MaterialRun]
            A representation of the material whose history is to be retrieved
        scope: Optional[str]
            [DEPRECATED] use a LinkByUID to specify a custom scope
            The scope of the uid. The lookup will be most efficient if you use the Citrine ID
            of the material, which is the default if scope=None.

        Returns
        -------
        MaterialRun
            A material run that has all of its fields fully populated with the processes,
            ingredients, measurements, and other materials that were involved in the history
            of the object.

        """
        link = _make_link_by_uid(id, scope)
        base_path = os.path.dirname(self._get_path(ignore_dataset=True))
        path = base_path + format_escaped_url("/material-history/{}/{}", link.scope, link.id)
        data = self.session.get_resource(path)

        # Add the root to the context and sort by writable order
        blob = dict()
        blob["context"] = sorted(
            data['context'] + [data['root']],
            key=lambda x: writable_sort_order(x["type"])
        )
        terminal_scope, terminal_id = next(iter(data['root']['uids'].items()))
        # Add a link to the root as the "object"
        blob["object"] = LinkByUID(scope=terminal_scope, id=terminal_id)

        # Serialize using normal json (with the GEMDEncoder) and then deserialize with the
        # GEMDEncoder encoder in order to rebuild the material history
        return MaterialRun.get_json_support().loads(
            json.dumps(blob, cls=GEMDEncoder, sort_keys=True))

    def get_by_process(self,
                       uid: Union[UUID, str, LinkByUID, GEMDProcessRun], *,
                       scope: Optional[str] = None) -> Optional[MaterialRun]:
        """
        [ALPHA] Get output material of a process.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDProcessRun]
            A representation of the process whose output is to be located.
        scope: Optional[str]
            [DEPRECATED] use a LinkByUID to specify a custom scope
            The scope of the uid, defaults to Citrine scope ("id")

        Returns
        -------
        MaterialRun
            The output material of the specified process, or None if no such material exists.

        """
        link = _make_link_by_uid(uid, scope)
        return next(
            self._get_relation(relation='process-runs', uid=link, per_page=1),
            None
        )

    def list_by_spec(self,
                     uid: Union[UUID, str, LinkByUID, GEMDMaterialSpec], *,
                     scope: Optional[str] = None) -> Iterator[MaterialRun]:
        """
        [ALPHA] Get the material runs using the specified material spec.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDMaterialSpec]
            A representation of the material spec whose material run usages are to be located.
        scope: Optional[str]
            [DEPRECATED] use a LinkByUID to specify a custom scope
            The scope of the uid, defaults to Citrine scope ("id")

        Returns
        -------
        Iterator[MaterialRun]
            The material runs using the specified material spec.

        """
        link = _make_link_by_uid(uid, scope)
        return self._get_relation('material-specs', uid=link)

    def list_by_template(self, uid: Union[UUID, str, LinkByUID, GEMDMaterialTemplate], *,
                         scope: Optional[str] = None) -> Iterator[MaterialRun]:
        """
        [ALPHA] Get the material runs using the specified material template.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDMaterialTemplate]
            A representation of the material template whose material run usages are to be located.
        scope: Optional[str]
            [DEPRECATED] use a LinkByUID to specify a custom scope
            The scope of the uid, defaults to Citrine scope ("id")

        Returns
        -------
        Iterator[MaterialRun]
            The material runs using the specified material template.

        """
        spec_collection = MaterialSpecCollection(self.project_id, self.dataset_id, self.session)
        specs = spec_collection.list_by_template(uid=_make_link_by_uid(uid, scope))
        return (run for runs in (self.list_by_spec(spec) for spec in specs)
                for run in runs)



from typing import List, Optional, Mapping, Union

from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import (
    FormulationDescriptor, RealDescriptor, CategoricalDescriptor
)
from citrine.informatics.predictors import Predictor

__all__ = ['MeanPropertyPredictor']


class MeanPropertyPredictor(
        VersionedEngineResource['MeanPropertyPredictor'], Predictor):
    """A predictor that computes a component-weighted mean of real or categorical properties.

    Each component in a formulation contributes to the mean property
    a weight equal to its quantity raised to the power `p`.
    For real-valued properties, the property values of each component are averaged
    with these weights to yield the component-weighted mean property.
    For categorical-valued properties, these weights are accumulated to yield a
    distribution over property values in the formulation.

    Parameters
    ----------
    name: str
        Name of the configuration
    description: str
        Description of the predictor
    input_descriptor: FormulationDescriptor
        Descriptor that represents the input formulation
    properties: List[Union[RealDescriptor, CategoricalDescriptor]]
        List of real or categorical descriptors to featurize
    p: float
        Power of the component-weighted mean.
        Positive, negative, and fractional powers are supported.
    impute_properties: bool
        Whether to impute missing ingredient properties.
        If ``False`` all ingredients must define values for all featurized properties.
        Otherwise, the row will not be featurized.
        If ``True`` and no ``default_properties`` are specified, then the average over the
        entire dataset is used.
        If ``True`` and a default is specified in ``default_properties``, then the specified
        default is used in place of missing values.
    label: Optional[str]
        Only ingredients with this label are counted when calculating the component-weighted mean.
        If ``None`` (default) all ingredients will be counted.
    default_properties: Optional[Mapping[str, Union[str, float]]]
        Default values to use for imputed properties.
        Defaults are specified as a map from descriptor key to its default value.
        If not specified and ``impute_properties == True`` the average over the entire dataset
        will be used to fill in missing values. Any specified defaults will be used in place of
        the average over the dataset. ``impute_properties`` must be ``True`` if
        ``default_properties`` are provided.
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. De-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    """

    input_descriptor = _properties.Object(FormulationDescriptor, 'data.instance.input')
    properties = _properties.List(
        _properties.Union(
            [_properties.Object(RealDescriptor), _properties.Object(CategoricalDescriptor)]
        ),
        'data.instance.properties'
    )
    p = _properties.Float('data.instance.p')
    impute_properties = _properties.Boolean('data.instance.impute_properties')
    label = _properties.Optional(_properties.String, 'data.instance.label')
    default_properties = _properties.Optional(
        _properties.Mapping(
            _properties.String,
            _properties.Union([_properties.String, _properties.Float])
        ),
        'data.instance.default_properties'
    )
    training_data = _properties.List(
        _properties.Object(DataSource), 'data.instance.training_data', default=[]
    )

    typ = _properties.String('data.instance.type', default='MeanProperty', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 input_descriptor: FormulationDescriptor,
                 properties: List[Union[RealDescriptor, CategoricalDescriptor]],
                 p: float,
                 impute_properties: bool,
                 label: Optional[str] = None,
                 default_properties: Optional[Mapping[str, Union[str, float]]] = None,
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description
        self.input_descriptor: FormulationDescriptor = input_descriptor
        self.properties: List[Union[RealDescriptor, CategoricalDescriptor]] = properties
        self.p: float = p
        self.impute_properties: bool = impute_properties
        self.label: Optional[str] = label
        self.default_properties: Optional[Mapping[str, Union[str, float]]] = default_properties
        self.training_data: List[DataSource] = training_data or []

    def __str__(self):
        return '<MeanPropertyPredictor {!r}>'.format(self.name)



"""
Citrine.io Python Client.

Documentation:
https://citrineinformatics.github.io/citrine-python/index.html

"""
from citrine.citrine import Citrine  # noqa: F401
import logging
import warnings
from .__version__ import __version__  # noqa: F401

logging.basicConfig(level=logging.WARNING)
warnings.simplefilter("default", DeprecationWarning)
